# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

from gettext import gettext as _

from dob_bright.termio import attr, fg, click_echo, echo_exit

from ..run_cli import run


def help_command_help(ctx, command=None):
    cmd = run
    if not command:
        # A simple `dob help`.
        cmd = run
        # The context is the 'help' command, which, if passed to
        # cmd.get_help, shows usage a bit incorrectly, as:
        #
        #   Usage: dob help [OPTIONS] COMMAND [ARGS]...
        #
        # (note the "help"), so specify the parent (the root 'dob'
        # command), so that usage gets printed more accurately as:
        #
        #   Usage: dob [OPTIONS] COMMAND [ARGS]...
        cmd_ctx = ctx.parent
    else:
        # For help on a subcommand, i.e., `dob help command`,
        # use the 'help' command to find the subcommand.
        cmd_ctx = ctx.parent
        parts = []
        for part in command:
            if part.startswith('-'):
                continue
            parts.append(part)
            cmd = cmd.get_command(cmd_ctx, part)
            if cmd is None:
                echo_exit(ctx, _(
                    """
For detailed help, try:
  {codehi}{helpcmd}{reset}

ERROR: No such command: “{command}”
                    """
                ).format(
                    command=' '.join(parts),
                    helpcmd=' '.join([ctx.command_path] + parts[:-1]),
                    # FIXME: (lb): Replace hardcoded. Assign from styles.conf. #styling
                    codehi=(fg('turquoise_2') or ''),
                    reset=(attr('reset') or ''),
                ).strip(), exitcode=1)
            if isinstance(cmd, tuple):
                cmd = cmd[1]
            # Make a temporary Context to print the help.
            # (lb): This couples dob to Click more tightly than I'd
            # like, but it also means I don't have to sell a lot of
            # little design tweaks to the Click team.
            args = []
            cmd_ctx = cmd.make_context(cmd.name, args, parent=cmd_ctx)

    click_echo(cmd.get_help(cmd_ctx))




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma,  2015-2016 Eric Goller.  All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

import os
import platform
import traceback

from gettext import gettext as _

from dob_bright.termio import (
    click_echo,
    dob_in_user_exit,
    dob_in_user_warning,
    highlight_value
)
from dob_bright.config.app_dirs import AppDirs, get_appdirs_subdir_file_path

from .echo_fact import write_fact_block_format
from .save_confirmed import prompt_and_save_confirmed


__all__ = (
    # (lb): aka 'prompt_and_save_paranoid'
    'prompt_and_save_backedup',
)


# ***

def prompt_and_save_backedup(
    controller,
    backup=True,
    leave_backup=False,
    rule='',
    dry=False,
    **kwargs,
):
    """"""

    def _prompt_and_save():
        backup_f = prepare_backup_file(backup)
        delete_backup = False
        inner_error = None
        saved_facts = []
        try:
            backup_callback = write_facts_file(backup_f, rule, dry)
            saved_facts = prompt_and_save_confirmed(
                controller,
                rule=rule,
                backup_callback=backup_callback,
                dry=dry,
                **kwargs,
            )
            delete_backup = True
        except SystemExit:
            # Explicit sys.exit() from our code.
            # The str(err) is just the exit code #.
            raise
        except BaseException as err:
            # NOTE: Using BaseException, not just Exception, so that we
            #       always catch (KeyboardInterrupt, SystemExit), etc.
            # Don't cleanup backup file.
            traceback.print_exc()
            inner_error = str(err)
        finally:
            if not delete_backup:
                traceback.print_exc()
                msg = 'Something horrible happened!'
                if inner_error is not None:
                    msg += _(' err: "{}"').format(inner_error)
                if backup_f:
                    msg += (
                        _("\nBut don't worry. A backup of edits was saved at: {}")
                        .format(backup_f.name)
                    )
                dob_in_user_exit(msg)
            cleanup_files(backup_f, delete_backup)
        return saved_facts

    # ***

    def prepare_backup_file(backup):
        if not backup:
            return None
        backup_path, backup_link = get_import_ephemeral_backup_path()
        log_msg = _("Creating backup at {0}").format(backup_path)
        controller.client_logger.info(log_msg)
        backup_f = backup_file_open(backup_path)
        backup_file_symlink(backup_path, backup_link)
        return backup_f

    def backup_file_open(backup_path):
        try:
            backup_f = open(backup_path, 'w')
        except Exception as err:
            msg = (
                'Failed to create temporary backup file at "{}": {}'
                .format(backup_path, str(err))
            )
            dob_in_user_exit(msg)
        return backup_f

    def backup_file_symlink(backup_path, backup_link):
        if platform.system() == 'Windows':
            # Windows only recently added symlinks, and even then, you need
            # privileges or to enable a special switch. But whatever, this
            # symlink is only a convenience. It's not necessary.
            return
        try:
            # NOTE: os.remove removes the file being linked; we want unlink.
            #   We also want lexists, not exists, to get True for broken links.
            os.unlink(backup_link) if os.path.lexists(backup_link) else None
            os.symlink(backup_path, backup_link)
        except Exception as err:
            msg = (
                'Failed to setup temporary backup file link at "{}": {}'
                .format(backup_link, str(err))
            )
            dob_in_user_warning(msg)

    IMPORT_BACKUP_DIR = 'carousel'

    def get_import_ephemeral_backup_path():
        backup_prefix = 'dob.import'
        backup_tstamp = controller.now.strftime('%Y%m%d%H%M%S')
        backup_basename = backup_prefix + '-' + backup_tstamp
        backup_fullpath = get_appdirs_subdir_file_path(
            file_basename=backup_basename,
            dir_dirname=IMPORT_BACKUP_DIR,
            appdirs_dir=AppDirs.user_cache_dir,
        )
        # 2018-06-29 18:56: This symlink really isn't that helpful...
        #   but we'll see if I start using it. At least for DEVing.
        backup_linkpath = get_appdirs_subdir_file_path(
            file_basename=backup_prefix,
            dir_dirname=IMPORT_BACKUP_DIR,
            appdirs_dir=AppDirs.user_cache_dir,
        )
        return backup_fullpath, backup_linkpath

    # ***

    def cleanup_files(backup_f, delete_backup):
        if not backup_f:
            return
        backup_f.close()
        if not delete_backup:
            return
        if not leave_backup:
            try:
                os.unlink(backup_f.name)
            except FileNotFoundError:
                # [lb]: 2019-01-17: Happening occasionally on dob-import
                # testing, not sure if related dev_breakpoint usage, or
                # not editing any Facts, or what, but not predictably
                # recreating. Not a problem if it fails, but why would it
                # fail, especially given that the code just called close().
                controller.client_logger.warning(
                    'nothing to cleanup?: backup file missing: {}'
                    .format(backup_f.name)
                )
        else:
            click_echo(
                _('Abandoned working backup at: {}')
                .format(highlight_value(backup_f.name))
            )

    # ***

    def write_facts_file(fact_f, rule, dry):
        def wrapper(carousel):
            if dry or not fact_f:
                return
            fact_f.truncate(0)
            # (lb): truncate doesn't move the pointer (you can peek()), and while
            # write seems to still work, it feels best to reset the pointer.
            fact_f.seek(0)
            for idx, fact in enumerate(carousel.prepared_facts):
                # The Carousel should only send us facts that need to be
                # stored, which excludes deleted Facts that were never stored.
                controller.affirm((not fact.deleted) or (fact.pk > 0))
                write_fact_block_format(fact_f, fact, rule, is_first_fact=(idx == 0))
            fact_f.flush()

        return wrapper

    # ***

    return _prompt_and_save()




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma,  2015-2016 Eric Goller.  All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""Module to provide u18n friendly help text strings for our CLI commands."""

from gettext import gettext as _

from dob_bright.config.fileboss import default_config_path, default_config_path_abbrev
from dob_bright.config.urable import ConfigUrable
from dob_bright.termio import attr, bg, coloring, fg, highlight_value

from ..copyright import assemble_copyright

from .. import __arg0name__, __package_name__

# Note that the help formatter reformats paragraphs when the help is
# displayed, to a width determined by the terminal and max_content_width.
# So the newlines you see in the text below will be removed and each
# paragraph will be reformatted with indents and new newlines inserted.
# - Your best bet to see how the text is formatted is to run the help.

# Note also that a lot of the help text is assembled at runtime, after
# dob has started and parsed command line arguments and configuration
# settings. This is because the global coloring() switch is off until
# after arguments are parsed and the config is read, and then it might
# be turned. But the way coloring works, ANSI is stripped when strings
# are assembled, as opposed to ANSI being stripped on output (the latter
# would make it easier to turn color on or off without having to rebuild
# strings, or to worry about when strings are generated; but our code is
# built the other way, so we have to wait for the coloring() question to
# be settled before building help text. Hence all the callbacks below.
#
# E.g./tl;dr, instead of seeing help text built like this:
#
#           SOME_HELP_TEXT = _("""Helps you.""")
#
#       you'll often see a builder method used instead, such as:
#
#           def SOME_HELP_TEXT(ctx):
#               return _("""Helps {color}you{reset}.""").format(...)

# ***
# *** [COMMON] Oft-used format() args.
# ***


def common_format():
    # (lb): Eh: This is a method because ENABLE_COLORS is False on startup,
    # and fg and attr will return empty strings if called when module is
    # sourced. So wait until help strings are built and it's known if color
    # or not.
    # FIXME: (lb): Replace hardcoded styles. Assign from styles.conf. #styling
    common_format = {
        'appname': highlight_value(__package_name__),
        'rawname': __package_name__,
        'codehi': fg('turquoise_2'),
        'reset': attr('reset'),
        'bold': attr('bold'),
        'italic': attr('italic'),
        'underlined': attr('underlined'),
        'wordhi': fg('chartreuse_3a'),
        'errcol': bg('red_1') + attr('bold'),
    }
    return common_format


# ***

def _command_create_force_help(what):
    return _(
        """
        If specified, overwrite {} if is exists.
        """
    ).format(what)


# ***
# *** [BARE] Command help.
# ***

def RUN_HELP_WHATIS():
    _help = _(
        """
        {appname} is a time tracker for the command line.
        """
        .strip()
    ).format(**common_format())
    return _help


def RUN_HELP_TRYME():
    _help = _(
        """
        - Try the demo to get acquainted with dob quickly,

          {codehi}{rawname} demo{reset}
        """
        .strip()
    ).format(**common_format())
    return _help


def RUN_HELP_QUICKIE():
    _help = _(
        """
        - To start dobbin immediately, init, then edit,

          {codehi}{rawname} init{reset}

          {codehi}{rawname} edit{reset}
        """
        .strip()
    ).format(**common_format())
    return _help


def RUN_HELP_HEADER():
    _help = _(
        """
        {what_is}

        {try_me}

        {quick_start}
        """
        .strip()
    ).format(
        what_is=RUN_HELP_WHATIS(),
        try_me=RUN_HELP_TRYME(),
        quick_start=RUN_HELP_QUICKIE(),
        **common_format()
    )
    return _help


def RUN_HELP_TLDR():
    _help = _(
        """
        {what_is}

        - To read lots more help, run the help command,

         \b
         {codehi}{rawname} help{reset}

        - To learn to dob quick and easy, try the demo,

         \b
         {codehi}{rawname} demo{reset}

        \b
        {copyright}
        """
        .strip()
    ).format(
        what_is=RUN_HELP_WHATIS(),
        try_me=RUN_HELP_TRYME(),
        copyright='\b\n        '.join(assemble_copyright()),
        **common_format()
    )
    return _help


def RUN_HELP_COMPLETE():
    _help = _(
        """
        {run_help_header}

        - Use {italic}help{reset} with any command to learn more, e.g.,

          {codehi}{rawname} init --help{reset}

          {codehi}{rawname} help init{reset}

          {codehi}{rawname} edit -h{reset}

        - Put global options {italic}before{reset} the command name, e.g.,

          {codehi}{rawname} --no-color details{reset}

        - Some commands have aliases, shown in (parentheses) below, e.g.,

          {codehi}{rawname} details{reset} {italic}is the same as{reset} {codehi}{rawname} info{reset}

        - For code rights and legal info, review the {copyrt_sym} and {scroll_sym},

          {codehi}{rawname} copyright{reset}

          {codehi}{rawname} license{reset}
        """  # noqa: E501 line too long
        .strip()
    ).format(
        run_help_header=RUN_HELP_HEADER(),
        copyrt_sym=coloring() and '©' or 'copyright',
        scroll_sym=coloring() and '📜' or 'license',
        **common_format()
    )
    return _help


def RUN_HELP_OVERVIEW(ctx):
    if (
        (ctx.command.name == 'run')
        and (
            ctx.invoked_subcommand
            or ctx.help_option_spotted
        )
    ):
        return RUN_HELP_COMPLETE()
    return RUN_HELP_TLDR()


# ***
# *** [HELP] Command help.
# ***

HELP_HELP = _(
    """
    Print help for the application, or for the specified command.
    """
)


# ***
# *** [VERSION] Command help.
# ***

VERSION_HELP = _(
    """
    Print the interface and library versions.
    """
)


# ***
# *** [LICENSE] Command help.
# ***

LICENSE_HELP = _(
    """
    Print the software license.
    """
)


# ***
# *** [LICENSE] Command help.
# ***

COPYRIGHT_HELP = _(
    """
    Print the software copyright.
    """
)


ABOUT_COMMAND_HELP = _(
    """
    Alias of 'copyright' command.
    """
)


# ***
# *** [DETAILS] Command help.
# ***

DETAILS_HELP = _(
    """
    Print details about the runtime environment.
    """
)


DETAILS_TMI_HELP = _(
    """
    Show AppDirs paths, too.
    """
)

# ***
# *** [ENVIRONS] Command help.
# ***

ENVIRONS_HELP = _(
    """
    Print shell-sourceable details about the runtime environment.

    Useful for setting up shell scripting, e.g.,

       \b
       eval "$({} environs)"
       echo $DOB_CONF
    """.format(__arg0name__)
)


# ***
# *** [DEBUG] Command help.
# ***

DEBUG_HELP = _(
    """
    Hidden command to break into a REPL prompt and poke around dob internals.

    This command is mostly an example to show developers how to easily debug.

    You'll probably want to just sprinkle your own breakpoints where you
    need them, e.g.,

      import sys, pdb; pdb.set_trace()

    or, if you're breaking into code after something stole the input,
    fix it first,

      import os, pdb; os.system("stty sane"); pdb.set_trace()
    """
)


# ***
# *** [DEMO] Command help.
# ***

def DEMO_HELP(ctx):
    _help = _(
        """
        Learn {rawname} -- {italic}Run this first!{reset}
        """
        .strip()
    ).format(**common_format())
    return _help


# ***
# *** [INIT] Command help.
# ***

def INIT_HELP_OVERVIEW(ctx):
    controller = ctx.obj

    _hint_sqlite = ''
    if controller.config['db.engine'] == 'sqlite':
        _hint_sqlite = _(
            """
        And if you know SQL, you can poke around the database file easily:

         \b
         {codehi}# Assuming sqlite3 is installed:{reset}
         {codehi}sqlite3 {cfg_db_path}{reset}

        But you'll probably just want to make sure you backup that file!
            """
            .strip()
        ).format(
            cfg_db_path=controller.config['db.path'],
            **common_format()
        )

    _help = _(
        """
        Create a default configuration file, and an empty database.

        - Unless it exists, init will create a default configuration at:

         \b
         {default_config_path}

        - Unless it exists, init will create an empty database file at:

         \b
         {hlg_db_path}

        After running init, you can see the contents of the config
        file by opening it in a text editor, or, better yet, you can
        dump it and get some more help with dob:

         \b
         {codehi}{rawname} config dump{reset}

        {_hint_sqlite}
        """
        .strip()
    ).format(
        default_config_path=highlight_value(default_config_path()),
        hlg_db_path=highlight_value(controller.config['db.path']),
        _hint_sqlite=_hint_sqlite,
        **common_format()
    )
    return _help


# ***
# *** [CONFIG] Commands help.
# ***

def CONFIG_GROUP_HELP(ctx):
    _help = _(
        """
        Manage user config settings (including editor key bindings).

        Some application behavior can be changed via config values.

        Config values can be persisted across invocation in a config
        file.

        You can also specify config values at runtime on the command line,
        or using environment variables.

        {underlined}Config File Location{reset}

        By default, {rawname} looks for a configuration file at:

            {default_config_path}

        - You can specify an alternative file location
        using the {codehi}{envkey}{reset} environment value.

          E.g., you could set the environment for just the command subshell like:

          \b
          {codehi}{envkey}=path/to/{rawname}.conf {rawname} COMMAND ...{reset}

          or you could export the environment variable first and then invoke {rawname}:

          \b
          {codehi}export {envkey}=path/to/{rawname}.conf{reset}
          {codehi}{rawname} COMMAND ...{reset}

        - You can alternatively specify the configuration file location
        using the {codehi}-F/--configfile{reset} global option, e.g.,
        using shorthand:

          \b
          {codehi}{rawname} -F path/to/{rawname}.conf COMMAND ...{reset}

          or using --option=value longhand:

          \b
          {codehi}{rawname} --configfile=path/to/{rawname}.conf COMMAND ...{reset}

        {underlined}Config Value Precedence{reset}

        - If no config value is specified for a setting, {rawname}
        uses a default value.

        - If a config value is found in the config value, that value
        takes precedence over the default value.

        - If a corresponding environment variable for the config value
        is found, that value is preferred over the value from the file.

          - The environment variable for each setting is formed from
        a prefix, {codehi}DOB_CONFIG_{reset}, followed by the uppercase
        section name and the uppercase setting name.

          E.g., here's how to specify the db.engine setting using its
        environment variable:

          \b
          {codehi}DOB_CONFIG_DB_ENGINE=sqlite {rawname} stats{reset}

        - If a config value is specified via the command line, that value is
        preferred over all other values.

          - You can specify config values using the {codehi}-C/--config{reset}
        option, e.g.,

          \b
          {codehi}{rawname} -C db.engine=sqlite stats{reset}

        {underlined}Config Command Overview{reset}

          - You can edit the config file directly, or (better yet) you can
        use the {codehi}dob set{reset} command to change its values.
        E.g., to enable coloring whenever you use {rawname}, run:

          \b
          {codehi}{rawname} config set client term_color True{reset}

          or omit the section name (because {rawname} is smart) and run instead:

          \b
          {codehi}{rawname} config set term_color True{reset}

          - If you remove the config file, or if you delete values from it,
        don't worry, {rawname} will use default values instead.

          - You can recreate the config file (and overwrite the existing file)
        with defaults by:

          \b
          {codehi}{rawname} config create --force{reset}

          - The best way to learn about all configurable settings is to
        print the config table, which includes a helpful message for each
        option:

          \b
          {codehi}{rawname} config show{reset}

          - If you think your config file is missing values, you can
        update it with missing settings by running the update command:

          \b
          {codehi}{rawname} config update{reset}

            - But you should not care about the contents of the config file
        if you stick to using {codehi}{rawname} config show{reset} and
        {codehi}{rawname} config set{reset} commands.

            - Although you might care about the config file contents if you'd
        like to add comments to it, which is supported.
        """
    ).format(
        default_config_path=highlight_value(default_config_path()),
        envkey=ConfigUrable.DOB_CONFIGFILE_ENVKEY,
        **common_format()
    )
    return _help


CONFIG_CREATE_HELP = _(
    """
    Write a new configuration file populated with default values.

    You can overwrite an existing configuration file using --force.
    """
)


CONFIG_CREATE_FORCE_HELP = _command_create_force_help(_('config file'))


CONFIG_EDIT_HELP = _(
    """
    Open the config file in your preferred $EDITOR.
    """
)


CONFIG_GET_HELP = _(
    """
    Print a configuration value from the config file.
    """
)


CONFIG_SET_HELP = _(
    """
    Write a configuration value to the config file.
    """
)


CONFIG_SHOW_HELP = _(
    """
    Print all config settings, including names, values, and help.
    """
)


CONFIG_UPDATE_HELP = _(
    """
    Write missing configuration values to the config file.
    """
)


# ***
# *** [STYLES] Commands help.
# ***

def STYLES_GROUP_HELP(ctx):
    _help = _(
        """
        Manage editor styles (centering, dimensions, coloring, and more).

        The easiest way to use styles is to choose one of the built-in
        styles and set it in your config.

        {underlined}Built-in Styles{reset}

        To view the list of styles, run:

          \b
          {codehi}{rawname} styles list{reset}

        {underlined}Activate a Style{reset}

        Assign the style you'd like to enable to `editor.styling`.

        For instance, to choose the 'night' style, which looks good on
        a dark background, run:

          \b
          {codehi}{rawname} config set editor styling night{reset}

        Or, to make the 'light' style active for a light background:

          \b
          {codehi}{rawname} config set editor styling light{reset}

        {underlined}Design your own Style{reset}

        dob styles are managed using a typical config file.

        Each section in the config file is one style.

        To enable a custom style, set editor.styling to its section
        name in the config file.

        To get started, create a styles file populated using a
        built-in style as a start.

        For example, to use the 'night' style as a base, run:

          \b
          {codehi}{rawname} styles create night{reset}

        Or, to start from scratch, use the 'default' style,
        or leave the style unspecified:

          \b
          {codehi}{rawname} styles create{reset}

        {underlined}Design your own Rules{reset}

        dob style rules are also managed using a config file.

        Each section in the rules file represents a single rule.

        Each rule defines one or more metadata values to match
        against your Facts, and if a rule matches, any styles
        it declares are applied to the interactive components.

        Each rule can also have a code block executed by eval.

        To generate a rules file with one example rule added,
        run:

          \b
          {codehi}{rawname} rules create{reset}

        {underlined}Edit Styles and Rules{reset}

        You can load the style file easily into your $EDITOR
        from dob:

          \b
          {codehi}{rawname} styles edit{reset}

        As well as the rules file:

          \b
          {codehi}{rawname} rules edit{reset}

        Or you could open them yourself. Their paths are listed
        in the config, e.g.:

          \b
          $ {codehi}{rawname} config get editor styles_fpath{reset}
          /home/user/.config/dob/styling/styles.conf

          \b
          $ {codehi}{rawname} config get editor rules_fpath{reset}
          /home/user/.config/dob/styling/rules.conf

        {underlined}Writing Styles and Rules{reset}

        Use the 'show' command to view the list of styles and rule
        settings and their descriptions:

          \b
          {codehi}{rawname} styles show{reset}

          \b
          {codehi}{rawname} rules show{reset}

        In addition to the styles and rules listed, you can also
        add additional settings.

        Each user-created style setting is named with a class name of
        your choosing, or it's named to match one of the special class
        names added automatically based on each Fact's metadata.

        The value for each user-created style setting is a space-separated
        list of class name of style strings, as recognized by Python Prompt
        Toolkit and Pygments.
        """
    ).format(
        **common_format()
    )
    return _help


STYLES_CREATE_HELP = _(
    """
    Write a new styles config file populated with basic styles options.

    You can overwrite an existing styles config file using --force.
    """
)


STYLES_CREATE_FORCE_HELP = _command_create_force_help(_('styles file'))


STYLES_CONF_HELP = _(
    """
    Print the styles file, a section from it, or an internal style.

    If no name is specified, prints the styles.conf file, in one exists.
    This includes any comments, too.

    If a name is specified, the matching section from the styles.conf
    file is printed, if one is found, including comments.

    Otherwise, dob looks for an internally-defined style with the
    matching name and prints that.

    On error, prints an error message.
    """
)


STYLES_EDIT_HELP = _(
    """
    Open the styles config file in your preferred $EDITOR.
    """
)


STYLES_LIST_HELP = _(
    """
    Display a list of built-in and user-declared style names.

    You can activate a style by assigning its value to `editor.styling`.

    For example, to use the 'night' style, run:

        $ dob config set editor styling night
    """
)


STYLES_SHOW_HELP = _(
    """
    View a table of basic style config settings and their descriptions.
    """
)


# ***
# *** [RULES] Commands help.
# ***

def RULES_GROUP_HELP(ctx):
    _help = _(
        """
        Manage styles (customize matching Facts, activities, tags, etc.).

        See the styles help for detailed information:

          \b
          {codehi}{rawname} styles --help{reset}
        """
    ).format(
        **common_format()
    )
    return _help


RULES_CREATE_HELP = _(
    """
    Write a new rules file populated with an example rule.

    You can overwrite an existing style rules file using --force.
    """
)


RULES_CREATE_FORCE_HELP = _command_create_force_help(_('style rules'))


RULES_CONF_HELP = _(
    """
    Print the rules file, or a single rule if a rule name is supplied.

    If no name is specified, prints the rules.conf file, in one exists.
    This includes any comments, too.

    If a name is specified, the matching section from the rules.conf
    file is printed, if one is found, including comments.

    On error, prints an error message.
    """
)


RULES_EDIT_HELP = _(
    """
    Open the rules file in your preferred $EDITOR.
    """
)


RULES_LIST_HELP = _(
    """
    Display a list of user-defined rule names.

    Each rule is enabled by default.

    To disable a rule, edit the rules file and
    set the rule's 'enabled' setting to False.

    Use the `rules conf` command to see the rules for each
    rule listed by this command, or use `styles conf <rule>`
    or `styles show <rule> to see the ruleset for a single
    rule.
    """
)


RULES_SHOW_HELP = _(
    """
    Show rule settings in a table, including name, default value,
    and description.

    If no name is supplied, an example ruleset is displayed.

    Otherwise, the named rule is fetched from the rules file
    and displayed.
    """
)


# ***
# *** [IGNORE] Commands help.
# ***

def IGNORE_GROUP_HELP(ctx):
    _help = _(
        """
        Manages ignore lists (to exclude matching names from prompts).
        """
    ).format(
        **common_format()
    )
    return _help


IGNORE_CREATE_HELP = _(
    """
    Write a new ignore file populated with example sections.

    You can overwrite an existing ignore file using --force.
    """
)


IGNORE_CREATE_FORCE_HELP = _command_create_force_help(_('ignore file'))


IGNORE_EDIT_HELP = _(
    """
    Open the ignore file in your preferred $EDITOR.
    """
)


IGNORE_LIST_HELP = _(
    """
    Display the list of ignore file section names.
    """
)


IGNORE_SHOW_HELP = _(
    """
    Show ignore rules in a table.
    """
)


# ***
# *** [STORE] Commands help.
# ***

STORE_GROUP_HELP = _(
    """
    Manage the database: create a new one; upgrade from Hamster; and more.
    """
)


STORE_CREATE_HELP = _(
    """
    Create an empty database file.

    You can overwrite an existing database file using --force.
    """
)


STORE_CREATE_FORCE_HELP = _command_create_force_help(_('data store'))


STORE_PATH_HELP = _(
    """
    Print the database path.
    """
)


STORE_URL_HELP = _(
    """
    Print the database URL.
    """
)


STORE_UPGRADE_LEGACY_HELP = _(
    """
    Migrate a legacy “Hamster” database to dob.
    """
)


STORE_UPGRADE_FORCE_HELP = _(
    """
    If specified, overwrite data store if is exists.
    """
)


# ***
# *** [STATS] Commands help.
# ***

# Print stats about your data (number of Facts, tags, etc.).
STATS_HELP = _(
    """
    Print stats about your data (Fact count, tag count, etc.).
    """
)


# ***
# *** [LIST] Commands help.
# ***

QUERY_GROUP_HELP = _(
    """
    Print activity, category, or tag {item_part}.
    """
)


QUERY_ITEM_HELP = _(
    """
    Print {item_type} {item_part}, optionally filtered and sorted.

    Finds all {item_types} by default, but results can be filtered by
    Fact start and end times and name matches.

    Results are sorted by name by default, but can be sorted
    by usage count, cumulative duration, or first Fact start
    time.

    Results are displayed in an ASCII table format by default,
    available in many different styles, or can be converted to
    comma-separated text (CSV), JSON, and other formats.

    SEARCH_TERM: Use to match specific {item_type} name(s), ORed together.
    """
)


LIST_GROUP_HELP = QUERY_GROUP_HELP.format(item_part=_('names'))


LIST_ACTIVITIES_HELP = QUERY_ITEM_HELP.format(
    item_type=_("activity"),
    item_types=_("activities"),
    item_part=_('names'),
)


LIST_CATEGORIES_HELP = QUERY_ITEM_HELP.format(
    item_type=_("category"),
    item_types=_("categories"),
    item_part=_('names'),
)


LIST_TAGS_HELP = QUERY_ITEM_HELP.format(
    item_type=_("tag"),
    item_types=_("tags"),
    item_part=_('names'),
)


# (lb): The `list facts` command is hidden, but supported for parity.
LIST_FACTS_HELP = QUERY_ITEM_HELP.format(
    item_type=_("fact"),
    item_types=_("facts"),
    item_part=_('names'),
)


# ***
# *** [FIND/SEARCH/JOURNAL/EXPORT] Commands help.
# ***

QUERY_HELP = _(
    """
Add a SEARCH_TERM to find any Facts with a description that contains
the specified search term. Or specify multiple terms to find any Fact
that contains one or more terms.

Use --broad to also loosely match the SEARCH_TERM against the activity,
category, and tag values.

Use --since and --until options to restrict the search to a specific
time range.

Use the --activity, --category, and --tags options to restrict the search
to exact activity, category, and/or tag names.

Use --order and --direction to specify how the results are sorted.

Use --limit and --offset to fetch a portion of the results.
    """
)


RESULTS_PROCESSING_HELP = _(
    """
Prints to stdout, or use --output to send to a file.

{query_help}

Combine --group-activity and --group-category to group by Activity@Category
names, or use them separately. You can also --group-tags. Or --group-days
to group by whole days. Use any combination of groupings to change the
results.

Use one or more --column options to choose which details are output,
and in what order.

To choose a different output format, use --format <format>, or one of
the format aliases (--journal, --json, etc.).

Some formats and output columns have their options, too, listed last
in the help.
    """
).format(query_help=QUERY_HELP)


SEARCH_HELP = _(
    """
Show all Facts, or those matching a search criteria.

{results_processing_help}
    """
).format(results_processing_help=RESULTS_PROCESSING_HELP)


def REPORT_HELP(ctx):
    _help = _(
        """
Generate a report on recent time usage.

This command is essentially an alias to:

  \b
  {codehi}{rawname} find --journal --since 'last week' \\{reset}
  {codehi}   --group-activity --group-category --group-days \\{reset}
  {codehi}   --sort usage --dir desc \\{reset}
  {codehi}   -l start_date -l duration -l sparkline -l actegory -l tags_freq{reset}

{results_processing_help}
        """
    ).format(results_processing_help=RESULTS_PROCESSING_HELP, **common_format())
    return _help


# ***
# *** [USAGE] Commands help.
# ***

USAGE_GROUP_HELP = QUERY_GROUP_HELP.format(item_part=_('usage'))


USAGE_ACTIVITIES_HELP = QUERY_ITEM_HELP.format(
    item_type=_("activity"),
    item_types=_("activities"),
    item_part=_('usage'),
)


USAGE_CATEGORIES_HELP = QUERY_ITEM_HELP.format(
    item_type=_("category"),
    item_types=_("categories"),
    item_part=_('usage'),
)


USAGE_TAGS_HELP = QUERY_ITEM_HELP.format(
    item_type=_("tag"),
    item_types=_("tags"),
    item_part=_('usage'),
)


# (lb): The `usage facts` command is hidden, but supported for parity.
USAGE_FACTS_HELP = QUERY_ITEM_HELP.format(
    item_type=_("fact"),
    item_types=_("facts"),
    item_part=_('usage'),
)


# ***
# *** [CURRENT-FACT] Commands help.
# ***

STOP_HELP = _(
    # Not DRY: Copied from first line of ADD_FACT_THEN.
    """
    Stop active Fact, ending it now or at the time specified.
    """
)


CANCEL_HELP = _(
    """
    Discard the active Fact.
    """
)


CURRENT_HELP = _(
    """
    Print the active Fact, if there is one.
    """
)


def NO_ACTIVE_FACT_HELP(ctx):
    _help = _(
        """
        No active Fact. Try {italic}starting{reset} a new Fact first.
        """
        .strip()
    ).format(**common_format())
    return _help


LATEST_HELP = _(
    """
    Print the latest completed Fact (with the most recent end time).
    """
)


HELP_CMD_SHOW = _(
    """
    Print the active Fact, if any, otherwise the latest Fact.
    """
)


# ***
# *** [ADD-FACT] Commands help.
# ***

def ADD_FACT_REFERRAL():
    _help = _(
        """
        For more help on this and the other add Fact commands, try

          {codehi}{rawname} --pager help add{reset}
        """
        .strip()
    ).format(**common_format())
    return _help


# verify_none
ADD_FACT_ON = _(
    """
    Alias for 'now' command, e.g., `dob on act@gory #tag: Blah...`.
    """
)


# verify_none
ADD_FACT_NOW = _(
    # FIXME/2019-11-22: (lb): I think "if nothing active" might be wrong.
    """
    Start a new Fact only if no Fact is active, using time now.
    """
)


# verify_none
ADD_FACT_START = _(
    # Not DRY: Copied from first line of ADD_FACT_AT.
    """
    Start a new Fact, beginning now or at the time specified.
    """
)


# verify_start
def ADD_FACT_AT(ctx):
    _help = _(
        """
        \b
        Start a new Fact, beginning now or at the time specified.

        \b
        This might stop the active Fact, if one exists, and the time
        specified comes after the active Fact start time.

        \b
        Or this might change the start and/or stop time of other Facts
        if the Fact being added overlaps other Facts' time windows.

        {}
        """
    ).format(ADD_FACT_REFERRAL())
    return _help


# verify_then
def ADD_FACT_THEN(ctx):
    # FIXME/2019-11-22: Verify the 'just use a colon' text, I think that works.
    _help = _(
        """
        Stop the active Fact, and start a new one, at given time or now.

        This is basically a shortcut for the at command, which requires a time
        offset, e.g.,

          {rawname} then Grinding beans for coffee.

        is the same as:

          {rawname} at +0: Grinding beans...

        If you want to specify an offset, you can, just use a colon,
        which could work well to throw down a gap Fact, e.g.,

          {rawname} then -5m: Woke up.
          {rawname} now Grinding beans...
        """
        .strip()
    ).format(**common_format())
    return _help


# verify_still
ADD_FACT_STILL = _(
    """
    Stop the active Fact, and start a new one, copying metadata.

    The new Fact is started with the same activity, category, and tags
    as the Fact that is ended.
    """
)


# Note that dob.transcode mentions a `since:` that's like `after:/next:` (verify_after).
# verify_after
def ADD_FACT_AFTER(ctx):
    _help = _(
        """
        Start a new Fact, beginning when the last Fact ended.

        {}
        """
        .strip()
    ).format(ADD_FACT_REFERRAL())
    return _help


# verify_after
ADD_FACT_NEXT = _(
    """
    Alias for the 'after' command, e.g., `dob next: Foo bar...`.
    """
)


# verify_end
ADD_FACT_TO = _(
    """
    Stop the active Fact, ending it now, or at the time specified.
    """
)


# verify_end
def ADD_FACT_UNTIL(ctx):
    _help = _(
        """
        Alias for the 'to' command, e.g., `{rawname} until -10m: Yada...`.
        """
        .strip()
    ).format(**common_format())
    return _help


# verify_both
def ADD_FACT_FROM(ctx):
    _help = _(
        """
        Insert a new Fact using the start and end time indicated.

        E.g., {rawname} from 2019-01-01 00:00 to 2019-01-01 01:00: Happy New Year!
        """
        .strip()
    ).format(**common_format())
    return _help


# ***
# *** [EDIT] Command help.
# ***

# (lb): 2019-11-22: I had been using the term "Carousel" in docs, but I
# think I should call it an "editor", and sometimes an "interactive" one.
EDIT_FACT_HELP = _(
    """
    Run the interactive editor in your terminal, to add and edit Facts.
    """
)


# ***
# *** [EXPORT] Command help.
# ***

def EXPORT_HELP(ctx):
    _help = _(
        """
Export Facts to a file, for backing up to text, or grepping.

{query_help}

See also {codehi}{rawname} report{reset} for more robust searching
and reporting options. The export command saves Facts to a file and
is useful for backing up your data as a text file, or for making a
searchable text file that you can wire into your existing notes
management system.
        """
    ).format(query_help=QUERY_HELP, **common_format())
    return _help


# ***
# *** [IMPORT] Command help.
# ***

# FIXME/2018-05-12: (lb): Document hamster-import more fully.
# - This command is quite powerful, and it might be a good way
# to demonstrate how all the dob-on/dob-after/etc. commands work.
IMPORT_HELP = _(
    """
Import Facts from a file, or from stdin, using a natural syntax.

Useful if you cannot use dob for a while, but you can maintain a
text file. Or if you need to massage data from another source into
dob.

HINT: To read from stdin, you can pipe to dob:

        echo "2020-01-09 00:00: Hi!" | dob import

      You can use shell redirection:

        dob import < path/to/my.facts

      Or you can use a single dash as the filename:

        dob import -

      For the last option, dob processes input as you type it,
      until you press ^D.
    """
)


# ***
# *** [COMPLETE] Command help.
# ***

COMPLETE_HELP = _(
    """
    Bash tab-completion helper.
    """
)


# ***
# *** [MIGRATE] Commands help.
# ***

MIGRATE_GROUP_HELP = _(
    """
    Upgrade the database, if necessary.
    """
)


MIGRATE_CONTROL_HELP = _(
    """
    Mark a database as under version control.
    """
)


MIGRATE_DOWN_HELP = _(
    """
    Downgrade the database version by 1 script.
    """
)


MIGRATE_UP_HELP = _(
    """
    Upgrade the database version by 1 script.
    """
)


MIGRATE_VERSION_HELP = _(
    """
    Show the database migration version.
    """
)


# ***
# *** [--GLOBAL-OPTIONS] Global Options help. [run_cli.py]
# ***

GLOBAL_OPT_VERBOSE = _(
    """
    Be chatty (use -VV for more).
    """
)


GLOBAL_OPT_VERBOSER = _(
    """
    Be chattier.
    """
)


GLOBAL_OPT_COLOR_NO_COLOR = _(
    """
    Color, or plain (Default: Auto).
    """
)


GLOBAL_OPT_PAGER_NO_PAGER = _(
    """
    Send output to pager, or not (Default: No).
    """
)


GLOBAL_OPT_CONFIG = _(
    """
    Override config setting(s) (multiple allowed).
    """
)


# MAYBE: (lb): Profiling: I tested this as a function, to avoid re.sub()
# always being called (on every invocation, in default_config_path_abbrev),
# but the option help is always built, so cannot escape. I wouldn't expect
# this to really have a performance impact, though.
GLOBAL_OPT_CONFIGFILE = _(
    """
    Specify config path (Default: {default_config_path_abbrev}).
    """
    .strip()
).format(
    default_config_path_abbrev=highlight_value(default_config_path_abbrev()),
)




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma,  2015-2016 Eric Goller.  All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""Copyright output UX methods."""

import os
from datetime import datetime

from gettext import gettext as _

from dob_bright.termio import click_echo

from . import (
    get_version,
    __arg0name__,
    __author_name__,
    __author_link__,
    __package_name__
)

__all__ = (
    'assemble_copyright',
    'echo_copyright',
    'echo_license',
)


def assemble_copyright():
    """Display a greeting message providing basic set of information."""
    cur_year = str(datetime.now().year)
    year_range = '2018'
    if cur_year != year_range:
        year_range = '2018-{}'.format(cur_year)
    gpl3_notice_2018 = [
        'This is {pkgname} v{version}.'.format(
            pkgname=__package_name__,
            version=get_version(),
        ),
        '',
        'Copyright (C) {years} {aname} <{alink}>'.format(
            years=year_range,
            aname=__author_name__,
            alink=__author_link__,
        ),
        # Be nice and call out the significant copyright holders from the years.
        # (lb): What about Right to be forgotten?
        'Copyright (C) 2015-2016 Eric Goller <elbenfreund@DenkenInEchtzeit.net>',
        '',
        _('This program comes with ABSOLUTELY NO WARRANTY. This is free software,'),
        _('and you are welcome to redistribute it under certain conditions.'),
        _(''),
        _('Run `{} license` for your legal rights and responsibilities.'
          ).format(__arg0name__),
    ]
    copy_notice = gpl3_notice_2018
    return copy_notice


def echo_copyright():
    notice = assemble_copyright()
    for line in notice:
        click_echo(line)


def echo_license():
    license = """
{app_name} is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

{app_name} is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
""".strip()
    license_txt = license.format(app_name=__package_name__)
    # MAYBE: (lb): Prefer the LICENSE file?
    # FIXME: LICENSE is probably not copied via pip-install.
    #        This will be good learning on installing non-package files.
    license_path = os.path.join(
        os.path.dirname(__file__),
        '..',
        'LICENSE',
    )
    if os.path.exists(license_path):
        with open(license_path, 'rb') as f:
            license_txt = f.read().decode('utf-8').strip()
    click_echo(license_txt)




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma,  2015-2016 Eric Goller.  All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

from gettext import gettext as _

from dob_bright.termio import (
    attr,
    click_echo,
    echo_block_header,
    highlight_value
)
from dob_bright.termio.crude_progress import CrudeProgress

from .echo_fact import echo_fact, write_fact_block_format
from .simple_prompts import mend_facts_confirm_and_save_maybe


__all__ = (
    'prompt_and_save_confirmer',
)


# ***

def prompt_and_save_confirmer(
    controller,
    edit_facts=None,
    orig_facts=None,
    backup_callback=None,
    file_out=None,
    rule='',
    yes=False,
    dry=False,
    progress=None,
    **kwargs,
):
    """"""
    progress = CrudeProgress(enabled=True) if progress is None else progress

    def _prompt_and_save():
        saved_facts = []
        try:
            saved_facts = persist_facts()
        finally:
            cleanup_files()
        return saved_facts

    def cleanup_files():
        # (lb): Click also closes; or we can hit it first.
        if file_out and not dry:
            file_out.close()

    # ***

    def persist_facts():
        if not edit_facts:
            return
        saved_facts = record_edited_facts()
        celebrate()
        return saved_facts

    def record_edited_facts():
        task_descrip = _('Saving facts')
        if progress is not None:
            term_width, dot_count, fact_sep = progress.start_crude_progressor(
                task_descrip,
            )

        new_and_edited = []

        other_edits = {fact.pk: fact for fact in edit_facts}

        for idx, fact in enumerate(edit_facts):
            if progress is not None:
                term_width, dot_count, fact_sep = progress.step_crude_progressor(
                    task_descrip, term_width, dot_count, fact_sep,
                )

            is_first_fact = idx == 0
            is_final_fact = idx == (len(edit_facts) - 1)
            fact_pk = fact.pk
            new_and_edited += persist_fact(
                fact, other_edits, is_first_fact, is_final_fact,
            )
            # If an existing Fact:
            #   - the pk is the same; and
            #   - the saved Fact is marked deleted, and a new one is created,
            #      or saved Fact is not marked deleted if it was ongoing Fact.
            # But if a new Fact, pk was < 0, now it's None, and new fact pk > 0.
            # Once saved, rely on Fact in store for checking conflicts.

            # If fact existed, fact.pk; else, fact_pk < 0 is in-app temp. ID.
            del other_edits[fact.pk is not None and fact.pk or fact_pk]

        assert len(other_edits) == 0

        # NOTE: It's up to the caller to ensure controller.post_process called,
        #       either via @post_processor, or by calling it directly.

        progress and progress.click_echo_current_task('')

        return new_and_edited

    def persist_fact(fact, other_edits, is_first_fact, is_final_fact):
        new_and_edited = [fact, ]
        if not dry:
            # If user did not specify an output file, save to database
            # (otherwise, we may have been maintaining a temporary file).
            if not file_out:
                new_and_edited = persist_fact_save(
                    fact, is_final_fact, other_edits,
                )
            else:
                write_fact_block_format(file_out, fact, rule, is_first_fact)
        else:
            echo_block_header(_('Fresh Fact!'))
            click_echo()
            echo_fact(fact)
            click_echo()
        return new_and_edited

    # ***

    def persist_fact_save(fact, is_final_fact, other_edits):
        # (lb): This is a rudimentary check. I'm guessing other code
        # will prevent user from editing old Fact and deleting its
        # end, either creating a second ongoing Fact, OR, more weirdly,
        # creating an ongoing Fact that has closed Facts after it!
        # 2018-07-05/TEST_ME: Test previous comment: try deleting end of old Fact.
        # FIXME: Do we care to confirm if is_final_fact is indeed latest ever? Meh?
        time_hint = 'verify_both' if not is_final_fact else 'verify_last'
        new_and_edited = mend_facts_confirm_and_save_maybe(
            controller, fact, time_hint, other_edits, yes=yes, dry=dry,
        )
        return new_and_edited

    # ***

    def celebrate():
        if not edit_facts:
            return
        click_echo('{}{}{}! {}'.format(
            attr('underlined'),
            _('Voilà'),
            attr('reset'),
            _('Saved {} facts.').format(highlight_value(len(edit_facts))),
        ))

    # ***

    return _prompt_and_save()




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma,  2015-2016 Eric Goller.  All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

from gettext import gettext as _

from ..clickux.query_assist import error_exit_no_results

from . import generate_usage_table

__all__ = ('usage_activities', )


def usage_activities(
    controller,
    show_usage=False,
    show_duration=False,
    output_format='table',
    table_type='texttable',
    max_width=-1,
    output_path=None,
    **kwargs
):
    """
    List all activities. Provide optional filtering by name.

    Args:
        search_term (str): String to match ``Activity.name`` against.

    Returns:
        None: If success.
    """
    def _usage_activities():
        err_context = _('activities')

        results = controller.activities.get_all_by_usage(**kwargs)

        results or error_exit_no_results(err_context)

        generate_usage_table(
            controller,
            results,
            name_header=_("Activity@Category"),
            name_fmttr=name_fmttr,
            show_usage=show_usage,
            show_duration=show_duration,
            output_format=output_format,
            table_type=table_type,
            max_width=max_width,
            output_path=output_path,
        )

    def name_fmttr(activity):
        if activity.category:
            category_name = activity.category.name
        else:
            category_name = '<NULL>'
        actegory = '{}@{}'.format(activity.name, category_name)
        return actegory

    # ***

    _usage_activities()




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma,  2015-2016 Eric Goller.  All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""``hamster``, ``hamster``, ``hamster``!!! a cuddly, furry time tracker."""

import os
import sys

from nark import get_version as _get_version

__all__ = (
    'get_version',
    '__arg0name__',
    '__author_name__',
    '__author_link__',
    '__package_name__',
)

__arg0name__ = os.path.basename(sys.argv[0])

# (lb): These are duplicated in setup.cfg:[metadata], but not sure how to DRY.
#   Fortunately, they're not likely to change.
__author_name__ = 'Landon Bouma'
__author_link__ = 'https://tallybark.com'

# (lb): Not sure if the package name is available at runtime. Seems kinda meta,
# anyway, like, Who am I? I just want to avoid hard coding this string in docs.
# (And it's also used for making the `dob version` output.)
__package_name__ = 'dob'


def get_version(include_head=False):
    return _get_version(
        package_name=__package_name__,
        reference_file=__file__,
        include_head=include_head,
    )




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

from functools import update_wrapper

import click_hotoffthehamster as click

from dob_bright.termio import click_echo, echo_exit

__all__ = (
    'show_help_finally',
    'show_help_if_no_command',
)


def show_help_finally(func):
    """Click command callback decorator to show help if requested.
    This changes the behavior of option parsing: The old behavior
    of `dob --help command` would show the general help, i.e., it
    would show the same help as `dob --help`. By telling Click to
    cool it, we can show the help for the specified sub-command
    instead, i.e., `dob --help command` will now show the help for
    command."""

    # (lb): I'll admit this might be the "hard way" to do this.
    # In run_cli.py, rather than passing help_option_names to
    # the Context, we could add our own click.option for -h/--help,
    # add a 'help' param to the run() method, and set the flag that
    # way. However, then each command would need to set the help
    # option, e.g., `dob --help command` sets help on the root Context
    # (via the run_cli.run() callback), whereas `dob command --help`
    # sets the help option on the command's context. Not that this
    # would be hard to do -- in addition to adding the show_help_finally
    # decorator to every Click command in dob.py, we could also add a
    # new click.option decorator to each of those commands... but then
    # we'd also have to add the 'help' param to every command method in
    # dob.py... and you see right there what a colossal pain using a
    # normal Click option to handle --help would be. So I guess it is
    # not the "hard way" having Click specially handle -h/--help.

    @click.pass_context
    def check_help(ctx, *args, **kwargs):
        if ctx.find_root().help_option_spotted:
            echo_exit(ctx, ctx.get_help())
        func(*args, **kwargs)

    return update_wrapper(check_help, func)


def show_help_if_no_command(func):
    @click.pass_context
    def show_help(ctx, *args, **kwargs):
        if ctx.invoked_subcommand is None:
            click_echo(ctx.command.get_help(ctx))
        func(*args, **kwargs)
    return update_wrapper(show_help, func)




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma,  2015-2016 Eric Goller.  All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

from gettext import gettext as _

import click_hotoffthehamster as click

from dob_bright.crud.fact_dressed import FactDressed
from dob_bright.crud.interrogate import ask_edit_with_editor
from dob_bright.termio import dob_in_user_exit, dob_in_user_warning

from dob_prompt.prompters.triple_prompter import ask_user_for_edits

from .save_backedup import prompt_and_save_backedup
from .simple_prompts import mend_facts_confirm_and_save_maybe

__all__ = ('edit_fact_by_pk', )


def edit_fact_by_pk(
    controller,
    key,
    use_carousel=True,
    edit_text=False,
    edit_meta=False,
):
    """"""
    def _edit_fact_by_pk():
        old_fact = fact_from_key(key)
        if old_fact is None:
            return warn_nothing_found(key)
        unedited = old_fact.copy() if old_fact else None
        # Pre-run Awesome Prompt and/or $EDITOR.
        # Then run the Interactive Editor Carousel, the star
        # of the show; or show the $EDITOR if nothing else done.
        # FIXME/2019-11-23: (lb): add_fact does $EDITOR, then Awesome.
        #                   - Which way is best? Make if configurable!!
        if edit_meta:
            _prompter = ask_user_for_edits(  # noqa: F841
                controller,
                # MAYBE: rename old_fact
                # fact=edit_fact,
                fact=old_fact,
                always_ask=True,
                # (lb): If we leave restrict_edit alone, the prompter will
                # prompt for act@gory, then tags, and description (show editor).
                # Otherwise, we could just ask for, e.g., act@gory, say:
                #   restrict_edit='actegory',
            )
            # FIXME: Missing diff with old Fact to see if edited (and tell user)
            #    and missing save Fact.
        if edit_text:
            _prompter = ask_user_for_edits(  # noqa: F841
                controller,
                # MAYBE: rename old_fact
                # fact=edit_fact,
                fact=old_fact,
                always_ask=True,
                restrict_edit='description',
            )
            # FIXME: Missing diff with old Fact to see if edited (and tell user)
            #    and missing save Fact.
        if use_carousel:
            edited_facts = edit_old_fact(old_fact)
        elif not edit_meta and not edit_text:
            edited_facts = edit_old_factoid(old_fact)
        elif old_fact != unedited:
            # So @post_processor get proper list of edited facts,
            # we have to check if really edited.
            edited_facts = [old_fact]
        else:
            edited_facts = []
        return edited_facts

    def fact_from_key(key):
        if key > 0:
            return fact_from_key_pk(key)
        return fact_from_key_relative_or_new(key)

    def fact_from_key_relative_or_new(key):
        old_fact = fact_from_key_relative(key)
        if old_fact or key > 0:
            return old_fact
        return FactDressed.new_gap_fact(start=controller.now)

    def fact_from_key_pk(key):
        try:
            old_fact = controller.facts.get(pk=key)
            return old_fact
        except KeyError:
            dob_in_user_exit(
                _("No fact found with ID “{0}”.").format(key)
            )

    def fact_from_key_relative(key):
        offset = -1 - key
        old_facts = controller.facts.get_all(
            sort_cols=('start',),
            sort_orders=('desc',),
            limit=1,
            offset=offset,
            deleted=False,
        )
        return old_facts[0] if old_facts else None

    def warn_nothing_found(key):
        if key > 0:
            msg = _('No Fact found with ID “{}”.'.format(key))
        else:
            msg = _('There are not that many Facts.')
        dob_in_user_warning(msg)
        return None

    # ***

    def edit_old_fact(old_fact):
        saved_facts = prompt_and_save_backedup(
            controller,
            edit_facts=[old_fact],
            # (lb): The whole point on dob-edit is to fire up the Carousel.
            use_carousel=True,
            # MEH/2019-02-01: (lb) Support --yes or --dry on dob-edit?
            yes=False,
            dry=False,
        )
        return saved_facts

    # ***

    def edit_old_factoid(old_fact):
        raw_fact = editor_interact(old_fact)
        time_hint = fact_time_hint(old_fact)
        new_fact = new_fact_from_factoid(raw_fact, old_fact, time_hint)
        echo_edited_fact(new_fact, old_fact)
        new_and_edited = confirm_and_save(new_fact, time_hint)
        return new_and_edited

    def editor_interact(old_fact):
        """Presents user with their EDITOR displaying the Factoid text.
        """
        # So that the user can edit act@gory and tags, and not solely
        # the description, marshal the Fact to a factoid string and
        # let the user edit that.
        old_raw_fact = old_fact.friendly_str(
            shellify=False,
            description_sep='\n\n',
            localize=True,
            colorful=False,
            show_elapsed=False,
        )

        # MAYBE/2020-01-28: Give user option of using dob-prompt to edit
        # act@gory and tags, then the description, e.g., something like
        # this?:
        #
        #   # MAYBE/2019-11-23: Call ask_user_for_edits instead?
        #                       What happens on active Fact that has body?
        #   __prompter = ask_user_for_edits(
        #       controller,
        #       fact=old_fact,
        #       always_ask=True,
        #       restrict_edit='description',
        #   )

        new_raw_fact = ask_edit_with_editor(controller, old_fact, old_raw_fact)

        if old_raw_fact == new_raw_fact:
            dob_in_user_exit(
                _("Nothing changed! New Fact same as the old Fact.")
            )

        return new_raw_fact

    def fact_time_hint(old_fact):
        if old_fact.start and old_fact.end:
            time_hint = 'verify_both'
        else:
            assert old_fact.start
            time_hint = 'verify_start'
        return time_hint

    def new_fact_from_factoid(raw_fact, old_fact, time_hint):
        # NOTE: Parser expects Iterable of input parts.
        new_fact, __err = old_fact.create_from_factoid(
            (raw_fact,),
            time_hint=time_hint,  # Either 'verify_both' or 'verify_start'.
            lenient=True,
        )
        # FIXME/2018-06-10: (lb): Do we care about the error?

        # (lb): Start with the PK of the old fact. On save/update, the
        # store will see that the fact already exists, so it'll make a
        # new copy of the fact (with a new ID), and mark the old fact
        # 'deleted' (thereby acting like a Wiki -- never delete!).
        new_fact.pk = old_fact.pk

        return new_fact

    def echo_edited_fact(new_fact, old_fact):
        if new_fact == old_fact:
            dob_in_user_exit(
                _("Nothing changed! New Fact same as old Fact.")
            )

        click.echo('An edited fact!: {}'.format(str(new_fact)))
        click.echo('\nThe diff!\n')
        click.echo(old_fact.friendly_diff(new_fact))

    def confirm_and_save(new_fact, time_hint):
        new_and_edited = mend_facts_confirm_and_save_maybe(
            controller,
            new_fact,
            time_hint,
            other_edits={},
            yes=False,
            dry=False,
        )
        return new_and_edited

        # ***

    return _edit_fact_by_pk()




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma,  2015-2016 Eric Goller.  All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

from dob import dob


class TestCommandCopyright(object):
    """Make shure our greeting function behaves as expected."""

    def test_shows_copyright(self, capsys):
        """Make sure we show basic copyright information."""
        dob.echo_copyright()
        out, err = capsys.readouterr()
        assert "Copyright" in out

    # Already tested, really; see: test_dob_license_wrapper (dob._license() test).
    def test_shows_license(self, capsys):
        """Make sure we display a brief reference to the license."""
        dob.echo_license()
        out, err = capsys.readouterr()
        assert 'GNU GENERAL PUBLIC LICENSE' in out
        assert 'Version 3' in out




# This file exists within 'dob-bright':
#
#   https://github.com/hotoffthehamster/dob-bright
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

import os
import pytest

from click_hotoffthehamster.testing import CliRunner

import dob.dob as dob  # for dob.run


@pytest.fixture
def runner(tmpdir):
    """Provide a convenient fixture to simulate execution of (sub-) commands."""
    def runner(args=[], keep_paths=False, **kwargs):
        # Override environments that AppDirs (thankfully) hooks. Ref:
        #   ~/.virtualenvs/dob/lib/python3.8/site-packages/appdirs.py

        # Override paths: (1) if caller running multiple command test
        # (keep_paths=True); or (2) if user wants theirs (DOB_KEEP_PATHS).
        if keep_paths or os.environ.get('DOB_KEEP_PATHS', False):
            XDG_CONFIG_HOME = os.environ['XDG_CONFIG_HOME']
            XDG_DATA_HOME = os.environ['XDG_DATA_HOME']
        else:
            path = tmpdir.strpath
            XDG_CONFIG_HOME = '{}/.config'.format(path)
            XDG_DATA_HOME = '{}/.local/share'.format(path)
        os.environ['XDG_CONFIG_HOME'] = XDG_CONFIG_HOME
        os.environ['XDG_DATA_HOME'] = XDG_DATA_HOME

        env = {
            'XDG_CONFIG_HOME': XDG_CONFIG_HOME,
            'XDG_DATA_HOME': XDG_DATA_HOME,
            # Do not overwrite ~/.cache/dob path, where dob.log lives,
            # so DEV tail sees test output, too:
            #   'XDG_CACHE_HOME': '{}/.cache'.format(path),
            # It should not be necessary to set the state directory:
            #   'XDG_STATE_HOME': '{}/.local/state'.format(path),
            # AppDirs also checks 2 other environs, generally set
            # to system paths, and also you'll find already existing
            # for your user, probably:
            #   'XDG_DATA_DIRS': '/usr/local/share' or '/usr/share',
            #   'XDG_CONFIG_DIRS': '/etc/xdg',
        }
        return CliRunner().invoke(dob.run, args, env=env, **kwargs)
    return runner




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma,  2015-2016 Eric Goller.  All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

import traceback

from gettext import gettext as _

import click_hotoffthehamster as click
from inflector import English, Inflector

from dob_bright.crud.fix_times import mend_fact_timey_wimey
from dob_bright.termio import (
    attr,
    click_echo,
    dob_in_user_exit,
    fg,
)

from .echo_fact import echo_fact

__all__ = (
    'mend_facts_confirm_and_save_maybe',
    # Private:
    #   'echo_ongoing_completed',
    #   'must_confirm_fact_edits',
    #   'save_facts_maybe',
)


# ***

def mend_facts_confirm_and_save_maybe(
    controller, fact, time_hint, other_edits, yes, dry,
):
    """"""
    def _mend_facts_confirm_and_save_maybe():
        new_fact_or_two, conflicts = mend_fact_timey_wimey(
            controller, fact, time_hint, other_edits,
        )
        saved_facts = confirm_conflicts_and_save_all(new_fact_or_two, conflicts)
        return saved_facts

    def confirm_conflicts_and_save_all(new_fact_or_two, conflicts):
        """"""
        # Ask user what to do about conflicts/edits.
        ignore_pks = other_edits.keys()
        must_confirm_fact_edits(controller, conflicts, yes, dry)
        saved_facts = save_facts_maybe(
            controller, new_fact_or_two, conflicts, ignore_pks, dry,
        )
        return saved_facts

    return _mend_facts_confirm_and_save_maybe()


# ***

def must_confirm_fact_edits(controller, conflicts, yes, dry):
    """"""

    def _must_confirm_fact_edits(conflicts, yes, dry):
        conflicts = cull_stopped_ongoing(conflicts)
        if not conflicts:
            return

        yes = yes or dry
        if not yes:
            echo_confirmation_banner(conflicts)

        n_conflict = 0
        n_confirms = 0
        for edited_fact, original in conflicts:
            n_conflict += 1
            if not yes:
                confirmed = echo_confirmation_edited(
                    n_conflict, edited_fact, original,
                )
                n_confirms += 1 if confirmed else 0
            else:
                controller.client_logger.debug(
                    _('Editing fact: {}').format(edited_fact)
                )

        if n_conflict != n_confirms:
            # (lb): This function is not used by the carousel -- only by
            # one-off CLI commands -- so blowing up here is perfectly fine.
            # (The carousel has its own error message display mechanism;
            #  and more importantly the carousel should never die,
            #  but should only ever be asked to die by the user.)
            dob_in_user_exit(_("Please try again."))

    def cull_stopped_ongoing(conflicts):
        return [con for con in conflicts if 'stopped' not in con[0].dirty_reasons]

    def echo_confirmation_banner(conflicts):
        # FIXME: (lb): Replace hardcoded styles. Assign from styles.conf. #styling
        click.echo()
        click.echo(_(
            'Found {}{}{} {}. '
            '{}Please confirm the following changes:{}'
            .format(
                fg('magenta'), len(conflicts), attr('reset'),
                Inflector(English).conditional_plural(len(conflicts), 'conflict'),
                attr('underlined'), attr('reset'),
            )
        ))

    def echo_confirmation_edited(n_conflict, edited_fact, original):
        click.echo()
        # MAYBE:
        #   from dob_viewer.ptkui.re_confirm import confirm
        #   confirmed = confirm()
        confirmed = click.confirm(
            text=_('Conflict #{}\n-----------\n{}\nReally edit fact?').format(
                n_conflict,
                original.friendly_diff(edited_fact),
            ),
            default=False,
            abort=False,
            # prompt_suffix=': ',
            # show_default=True,
            # err=False,
        )
        return confirmed

    # ***

    _must_confirm_fact_edits(conflicts, yes, dry)


# ***

def save_facts_maybe(controller, new_facts, conflicts, ignore_pks, dry):
    """"""

    def _save_facts_maybe(controller, new_facts, conflicts, ignore_pks, dry):
        new_and_edited = []
        if conflicts and dry:
            echo_dry_run()
        for edited_fact, original in conflicts:
            if not dry:
                new_and_edited += save_fact(controller, edited_fact, dry)
                if 'stopped' in edited_fact.dirty_reasons:
                    # This'll happen on one-off dob-to/dob-at/etc., but it
                    # will not happen on dob-import, e.g., if dob-import
                    # closes ongoing fact, it'll be saved normally, and it
                    # will not be passed herein as part of conflicts.
                    echo_ongoing_completed(controller, edited_fact)
            else:
                click.echo(original.friendly_diff(edited_fact))

        for fact in new_facts:
            new_and_edited += save_fact(controller, fact, dry, ignore_pks=ignore_pks)
        return new_and_edited

    def echo_dry_run():
        click.echo()
        click.echo('{}Dry run! These facts will be edited{}:\n '.format(
            attr('underlined'),
            attr('reset'),
        ))

    def save_fact(controller, fact, dry, ignore_pks=[]):
        # (lb): SIMILAR: edits_manager.save_edited_fact, create.save_fact.
        if fact.pk and fact.pk < 0:
            fact.pk = None
        if fact.pk is None and fact.deleted:
            controller.client_logger.debug('{}: {}'.format(_('Dead fact'), fact.short))
            return []
        if not dry:
            controller.client_logger.debug('{}: {}'.format(_('Save fact'), fact.short))
            try:
                new_fact = controller.facts.save(fact, ignore_pks=ignore_pks)
            except Exception as err:
                traceback.print_exc()
                dob_in_user_exit(str(err))
        else:
            new_fact = fact
            echo_fact(fact)
        return [new_fact, ]

    # ***

    return _save_facts_maybe(controller, new_facts, conflicts, ignore_pks, dry)


# ***

def echo_ongoing_completed(controller, fact, cancelled=False):
    """"""
    def _echo_ongoing_completed():
        colorful = controller.config['term.use_color']
        leader = _('Completed: ') if not cancelled else _('Cancelled: ')
        cut_width = width_avail(len(leader))
        completed_msg = echo_fact(leader, colorful, cut_width)
        controller.client_logger.debug(completed_msg)

    def width_avail(leader_len):
        term_width = click.get_terminal_size()[0]
        width_avail = term_width - leader_len
        return width_avail

    def echo_fact(leader, colorful, cut_width):
        completed_msg = (
            leader
            + fact.friendly_str(
                shellify=False,
                description_sep=': ',

                # FIXME: (lb): Implement localize.
                # FIXME/2018-06-10: (lb): fact being saved as UTC
                localize=True,

                colorful=colorful,
                cut_width_complete=cut_width,
                show_elapsed=True,
            )
        )
        click_echo(completed_msg)
        return completed_msg

    _echo_ongoing_completed()




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""Click Group wrapper adds plugin support."""

from functools import update_wrapper
import glob
import os

from gettext import gettext as _

import click_hotoffthehamster as click

from dob_bright.config.app_dirs import AppDirs
from dob_bright.termio import dob_in_user_warning

from ..helpers.path import compile_and_eval_source

__all__ = (
    'ensure_plugged_in',
    'ClickPluginGroup',
    'PLUGINS_DIRNAME',
)


PLUGINS_DIRNAME = 'plugins'


class ClickPluginGroup(click.Group):

    def __init__(self, *args, **kwargs):
        super(ClickPluginGroup, self).__init__(*args, **kwargs)
        self.plugins_basepath = os.path.join(
            AppDirs.user_config_dir, PLUGINS_DIRNAME,
        )
        self.has_loaded = False

    @property
    def plugin_paths(self):
        py_paths = glob.glob(os.path.join(self.plugins_basepath, '*.py'))
        return py_paths

    def list_commands(self, ctx):
        """Return list of commands."""
        set_names = set()
        for cmd in self.get_commands_from_plugins(ctx, name=None):
            set_names.add(cmd.name)
        cmd_names = super(ClickPluginGroup, self).list_commands(ctx)
        return cmd_names

    def get_command(self, ctx, name):
        # Aha!:
        #   assert ctx.command is self  # So True.
        # NOTE: get_command is called via self.resolve_command, from
        #       click.MultiCommand.invoke. Just FYI. -(lb)
        # Call the get-commands func., which really just sources the plugins, so they
        # can tie into Click; then we can just call the base class implementation.
        cmd = super(ClickPluginGroup, self).get_command(ctx, name)
        if cmd is None:
            # (lb): Profiling: Loading plugins [2018-07-15: I have 3]: 0.139 secs.
            #       So only call if necessary.
            self.get_commands_from_plugins(ctx, name)
            cmd = super(ClickPluginGroup, self).get_command(ctx, name)
        return cmd

    def ensure_plugged_in(self, controller):
        if self.has_loaded:
            return
        self.get_commands_from_plugins(controller.ctx, name=None)
        # Redo the config now that the plugins are loaded (because when we
        # first read the config, the plugins had not added their definitions
        # yet, so any user plugin config was previously ignored).
        controller.replay_config()

    def get_commands_from_plugins(self, ctx, name):
        cmds = set()
        for py_path in self.plugin_paths:
            try:
                files_cmds = self.open_source_eval_and_poke_around(py_path, name)
                if files_cmds:
                    # (lb): We use a set, because if different plugins all import
                    # the same object, e.g., `from dob.run_cli import run`, we do
                    # not want to return multiple matches.
                    cmds |= files_cmds
            except Exception as err:
                msg = _(
                    'ERROR: Could not open plugins file "{}": {}'
                ).format(py_path, str(err))
                dob_in_user_warning(msg)
        self.has_loaded = True
        return list(cmds)

    def open_source_eval_and_poke_around(self, py_path, name):
        # NOTE: The code that's eval()ed might append to self._aliases!
        #       (Or anything else!)
        # NOTE: This source *should* be trusted -- the user had to run
        #       `dob plugin install` to wire it. At least I think so. -lb.
        eval_globals = compile_and_eval_source(py_path)
        cmds = self.probe_source_for_commands(eval_globals, name)
        return cmds

    def probe_source_for_commands(self, eval_globals, name):
        # Check for alias now, after having sourced the plugin.
        cmd_name = None
        if name:
            cmd_name = self.resolve_alias(name)

        cmds = set()
        for lname, obj in eval_globals.items():
            if not isinstance(obj, click.Command):
                continue
            if self is obj:
                # This is the 'run' object created in the run_cli module.
                continue
            if not cmd_name or obj.name == cmd_name:
                cmds.add(obj)
        return cmds


# ***

def ensure_plugged_in(func):
    """
    """

    def wrapper(ctx, controller, *args, **kwargs):
        # Ensure plugins are loaded for Click functions decorated like:
        #   @pass_controller
        #   @click.pass_context
        #   @ensure_plugged_in
        # - ctx.parent is a click_hotoffthehamster.core.Context, and
        #   ctx.parent.command is <ClickAliasableBunchyPluginGroup run>.
        ctx.parent.command.ensure_plugged_in(controller)
        func(ctx, controller, *args, **kwargs)

    return update_wrapper(wrapper, func)




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma,  2015-2016 Eric Goller.  All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""
Fixtures available to the tests/.

- In general, fixtures should return a single instance.

- If a fixture is a factory, its name should reflect that.

- A fixture that is parametrized should be suffixed with
  ``_parametrized`` to imply it has increased complexity.
"""

import pytest

from dob_bright.crud.fact_dressed import FactDressed

# We leave this conftest out of pytest_plugins, otherwise its
# test_fact_cls fixture is not overridden by the one below.
# (lb): I tested a workaround, putting the test_fact_cls below in
# another module, which I then listed in pytest_plugins, trying
# it both before and after the other one ("dob_bright.tests.conftest"),
# but the other conftest's fixture always wins (so maybe pytest sorts
# pytest_plugins?).
from dob_bright.tests.conftest import *  # noqa: F401, F403

# Load all upstream fixtures into the test namespace, as
# though the fixtures from dob-bright were defined herein.
# - It's possible to import the fixtures with a *-glob, e.g.,
#     from dob_bright.tests.conftest import *  # noqa: F401, F403
#   but using the pytest_plugins mechanism is more proper.
#   Ref:
#     https://docs.pytest.org/en/2.7.3/plugins.html
pytest_plugins = (
    "nark.tests.backends.sqlalchemy.conftest",
    # Make sure fixtures required by fixtures available, e.g., 'base_config'.
    "nark.tests.conftest",
    # SKIP (see comment, below):
    #    "dob_bright.tests.conftest",
    "tests.cli_runner",
    "tests.dob_runner",
)


# ***

@pytest.fixture
def test_fact_cls():
    return FactDressed


@pytest.fixture(scope="session")
def test_fact_cls_ro():
    return FactDressed


# ***

# (lb): Possible scope values: function, class, module, package or session.
# - Let's try the broadest scope. This is just for testing reports.
#   It drastically improves runtime at the expense of always testing
#   the same data store input for each test that uses this fixture.
#   - But we test enough using item factories in 'function'-scoped tests
#     that we should be okay.
@pytest.fixture(scope="session")
def five_report_facts_ctl(
    alchemy_store_ro,
    test_fact_cls_ro,
    set_of_alchemy_facts_ro,
    controller_with_logging_ro,
):
    controller = controller_with_logging_ro
    # You'll see 5 Facts in the store:
    #   alchemy_store_ro.session.execute('SELECT COUNT(*) FROM facts;').fetchall()
    controller.store = alchemy_store_ro
    controller.store.fact_cls = test_fact_cls_ro  # FactDressed
    # Ignored: set_of_alchemy_facts_ro.
    yield controller




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma,  2015-2016 Eric Goller.  All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""A time tracker for the command line. Utilizing the power of nark."""

from dob_bright.crud.parse_input import parse_input
from dob_bright.termio.crude_progress import CrudeProgress

from .save_backedup import prompt_and_save_backedup


__all__ = (
    'import_facts',
)


def import_facts(
    controller,
    file_in=None,
    file_out=None,
    rule='',
    backup=True,
    leave_backup=False,
    use_carousel=False,
    dry=False,
    **kwargs,
):
    """
    Import Facts from STDIN or a file.
    """

    # Bah. Progress is only useful if mend_facts_times calls insert_forcefully,
    # otherwise the import operation should be fast (at least for the 1Kish Fact
    # imports this author has ran). And insert_forcefully is only needed if the
    # Fact range being imported overlaps with existing Facts. Which it really
    # shouldn't, i.e., the use case is, I've been on vacation and Hamstering to
    # a file on my phone, and now I want that data imported into Hamster, which
    # will be strictly following the latest Fact saved in the store.
    progress = CrudeProgress(enabled=True)

    def _import_facts():
        new_facts = parse_input(
            controller,
            file_in=file_in,
            progress=progress,
        )
        saved_facts = prompt_and_save_backedup(
            controller,
            edit_facts=new_facts,
            file_out=file_out,
            rule=rule,
            backup=backup,
            leave_backup=leave_backup,
            use_carousel=use_carousel,
            dry=dry,
            yes=False,
            progress=progress,
            **kwargs,
        )
        return saved_facts

    # ***

    return _import_facts()




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma,  2015-2016 Eric Goller.  All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

from gettext import gettext as _

import click_hotoffthehamster as click

from .simple_prompts import echo_ongoing_completed

__all__ = (
    'cancel_fact',
)


def cancel_fact(controller, purge=False):
    """
    Cancel current Fact, either marking it deleted, or really removing it.

    Returns:
        None: If success.

    Raises:
        KeyErŕor: No active Fact can be found.
    """
    try:
        fact = controller.facts.cancel_current_fact(purge=purge)
    except KeyError:
        message = _("Nothing tracked right now. Not doing anything.")
        controller.client_logger.info(message)
        raise click.ClickException(message)
    else:
        echo_ongoing_completed(controller, fact, cancelled=True)
        return fact




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""Click Group wrapper combines all our custom classes."""

from gettext import gettext as _

import click_hotoffthehamster as click
from click_hotoffthehamster.core import Option

from dob_bright.termio.style import attr

from ._ansitextwrap import AnsiTextWrapper

__all__ = (
    'ClickBetterUsageGroup'
)


class ClickBetterUsageGroup(click.Group):

    # (lb): What I originally had but pretty wordy,
    # and tends to force usage to wrap:
    # USAGE_HINT_ALL_OPTS = '--GLOBAL-OPTIONS...'
    # USAGE_HINT_CMD_OPTS = '--COMMAND-OPTIONS...'

    USAGE_HINT_ALL_OPTS = '--GLOBAL...'
    USAGE_HINT_CMD_OPTS = '--OPTION...'
    USAGE_HINT_CMD_ARGS = 'ARGS...'

    def __init__(self, *args, **kwargs):
        super(ClickBetterUsageGroup, self).__init__(*args, **kwargs)

    def help_usage_command_path_postfix(self, ctx):
        if self.name == 'run':
            return ''
        # Return everything but the program name, e.g., if ctx.command_path
        # is "dob config dump", returns "config dump".
        return ctx.command_path.split(' ', 1)[1]

    def format_usage(self, ctx, formatter):
        """Writes the usage line into the formatter.

        This is a low-level method called by :meth:`get_usage`.

        Overriden by dob to emphasize with _underline_ and **bold**.
        """
        # We want usage to appear like "dob [--GLOBAL-OPTIONS...] init...",
        # but default click uses ctx.command_path, which is, e.g., "dob init".
        if self.name == 'run':
            prog = ctx.command_path
        else:
            prog = ctx.command_path.split(' ', 1)[0]
        prog = '{} [{}]'.format(
            prog, ClickBetterUsageGroup.USAGE_HINT_ALL_OPTS,
        )
        commands = self.help_usage_command_path_postfix(ctx)
        if commands:
            prog = '{} {}'.format(prog, commands)
        prog = '{bold}{prog}{reset}'.format(
            bold=attr('bold'),
            prog=prog,
            reset=attr('reset'),
        )
        #
        pieces = self.collect_usage_pieces(ctx)
        args = ''
        if pieces:
            args = '{bold}{args}{reset}'.format(
                bold=attr('bold'),
                args=' '.join(pieces),
                reset=attr('reset'),
            )
        #
        prefix = '{underlined}{usage}{reset}: '.format(
            underlined=attr('underlined'),
            usage=_('Usage'),
            reset=attr('reset'),
        )
        #
        formatter.write_usage(
            prog,
            args,
            prefix=prefix,
            cls=AnsiTextWrapper,
            alt_fmt=True,
        )

    def collect_usage_pieces(self, *args, **kwargs):
        """Show '[OPTIONS]' in usage unless command takes none."""
        # MAYBE/2019-11-15: Move this up into Click. Seems legit.
        n_opts = 0
        for param in self.params:
            if isinstance(param, Option):
                n_opts += 1
            # (lb): Skip click.core.Argument objects, which are
            # added separately to the pieces collection.
        self.options_metavar = ''
        if n_opts and self.name != 'run':
            # E.g., [--COMMAND-OPTIONS...]
            self.options_metavar = '[{}]'.format(
                ClickBetterUsageGroup.USAGE_HINT_CMD_OPTS,
            )
        # else, if n_opts and name == 'run', then --GLOBAL-OPTIONS shown instead.
        self.subcommand_metavar = ''
        if self.commands:
            self.subcommand_metavar = 'COMMAND [{}] [{}]'.format(
                ClickBetterUsageGroup.USAGE_HINT_CMD_OPTS,
                ClickBetterUsageGroup.USAGE_HINT_CMD_ARGS,
            )
        return super(
            ClickBetterUsageGroup, self
        ).collect_usage_pieces(*args, **kwargs)

    def format_help(self, ctx, formatter):
        """Override Click.Command: shorter help for commandless root command action."""
        self.format_usage(ctx, formatter)
        self.format_help_text(ctx, formatter)
        if (
            (ctx.command.name == 'run')
            and (ctx.invoked_subcommand is None)
            and (not ctx.help_option_spotted)
        ):
            return
        self.format_options(ctx, formatter)
        self.format_epilog(ctx, formatter)




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""Click Group wrapper adds plugin support."""

from collections import OrderedDict

import click_hotoffthehamster as click
from click_hotoffthehamster._compat import term_len

__all__ = (
    'ClickBunchyGroup',
)


class ClickBunchyGroup(click.Group):

    def __init__(self, *args, help_weight=10, **kwargs):
        super(ClickBunchyGroup, self).__init__(*args, **kwargs)
        self.group_bunchies = {None: OrderedDict()}
        self.order_sortkeys = {None: 0}
        self.help_weight = help_weight

    def add_command(self, cmd, name=None):
        self.group_bunchies[None][cmd.name] = None
        return super(ClickBunchyGroup, self).add_command(cmd, name)

    def add_to_bunch(self, cmd, bunchy_name, sort_key=0):
        self.bunchies_regroup(cmd, bunchy_name)
        self.sortkeys_update(cmd, bunchy_name, sort_key)

    def bunchies_regroup(self, cmd, bunchy_name):
        self.group_bunchies.setdefault(bunchy_name, OrderedDict())
        self.group_bunchies[bunchy_name][cmd.name] = None
        del self.group_bunchies[None][cmd.name]

    def sortkeys_update(self, cmd, bunchy_name, sort_key):
        self.order_sortkeys[bunchy_name] = sort_key

    # SYNC_ME: (lb): This is nasty...
    # Override Click's MultiCommand implementation.
    def format_commands(self, ctx, formatter):
        # Avoid list mutation when iterating and call list_commands now, so that
        # the plugins are sourced. (I.e., if a plugin calls add_to_bunch, it edits
        # group_bunchies, which should not (cannot) happen while under iteration.)
        _ignored = self.list_commands(ctx)  # noqa: F841 local var ... never used

        for tup in sorted(self.order_sortkeys.items(), key=lambda tup: tup[1]):
            commands, col_max = self.format_commands_fetch(ctx, tup[0])
            section_header = callable(tup[0]) and tup[0]() or tup[0]
            self.format_commands_write(
                commands, ctx, formatter, section_header, col_min=col_max
            )

    # Override Click's MultiCommand implementation.
    def format_commands_fetch(self, ctx, bunchy_name):
        commands = []
        col_max = 0
        _commands = super(ClickBunchyGroup, self).format_commands_fetch(ctx)
        for subcommand, cmd in _commands:
            # GROUP-BUNCH: (lb): A little wonky. Seems least disruptive to code, though,
            # i.e., this is a lazy approach to solving group command bunching.
            if subcommand in self._commands:
                aliases = '|'.join(sorted(self._commands[subcommand]))
                subcommand = '{0} ({1})'.format(subcommand, aliases)
            col_max = max(col_max, term_len(subcommand))
            # Note that this width does not account for color (ANSI codes).
            if cmd.name not in self.group_bunchies[bunchy_name].keys():
                continue
            commands.append((subcommand, cmd))
        # Sort commands in each group by help_weight, then name.
        sorted_commands = sorted(commands, key=lambda tup: tup[1].name)
        sorted_commands = sorted(sorted_commands, key=lambda tup: tup[1].help_weight)
        return sorted_commands, col_max




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma,  2015-2016 Eric Goller.  All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

import random

from gettext import gettext as _

from dob_bright.crud.fact_from_factoid import must_create_fact_from_factoid
from dob_bright.crud.fix_times import mend_fact_timey_wimey
from dob_bright.termio import dob_in_user_exit

from dob_prompt.prompters.triple_prompter import ask_user_for_edits

from .save_backedup import prompt_and_save_backedup

__all__ = (
    'add_fact',
)


# ***

def add_fact(
    controller,
    factoid,
    time_hint,
    use_carousel=False,
    edit_text=False,
    edit_meta=False,
    yes=False,
    dry=False,
):
    """
    Start or add a fact.

    Args:
        factoid (str): ``factoid`` detailing Fact to be added.
            See elsewhere for the factoid format.

        time_hint (str, optional): One of:
            | 'verify_none': Do not expect to find any time encoded in factoid.
            | 'verify_both': Expect to find both start and end times.
            | 'verify_start': Expect to find just one time, which is the start.
            | 'verify_end': Expect to find just one time, which is the end.
            | 'verify_then': Optional time is new start; and either extend
            ongoing fact to new start, or back-fill interval gap.
            | 'verify_still': Optional time is new start; copy prev meta to new Fact;
            either extend ongoing fact, or back-fill interval gap.
            | 'verify_after': No time spec. Start new Fact at time of previous end.

        yes (bool, optional): If True, update other Facts changed by the new
            fact being added (affects other Facts' start/end/deleted attrs).
            If False, prompt user (i.e., using fancy interface built with
            python-prompt-toolkit) for each conflict.

        edit_meta (bool, optional): If True, prompt user for activity and/or
            category, if not indicated; and prompt user for tags. Shows
            MRU lists to try to make it easy for user to specify commonly
            used items.

    Returns:
        Nothing: If everything went alright. (Otherwise, will have exited.)
    """

    def _add_fact():
        new_fact = _create_fact()
        new_fact_or_two, conflicts = _mend_times(new_fact)
        edit_facts, orig_facts = _prepare_facts(new_fact_or_two)
        edit_fact = _add_conflicts(conflicts, edit_facts, orig_facts)
        _maybe_prompt_description(edit_fact)
        _maybe_prompt_actegory(edit_fact)
        saved_facts = _prompt_and_save(edit_facts, orig_facts)
        return saved_facts

    def _create_fact():
        # NOTE: factoid is an ordered tuple of args; e.g., sys.argv[2:], if
        #       sys.argv[0] is the executable name, and sys.argv[1] is command.
        #       Because dob is totes legit, it does not insist that the user
        #       necessary quote anything on the command line, but it does not
        #       insist not, either, so factoid might have just one element (a
        #       string) containing all the information; or factoid might be a
        #       list of strings that need to be parsed and reassembled (though
        #       not necessarily in that order).
        #
        #         tl;dr: factoid is a tuple of 1+ strings that together specify the Fact.

        # Make a new Fact from the command line input.
        new_fact = must_create_fact_from_factoid(
            controller, factoid, time_hint,
        )
        return new_fact

    def _mend_times(new_fact):
        # If there's an ongoing Fact, we might extend or squash it.
        # Also, if the new Fact overlaps existing Facts, those Facts'
        # times might be changed, and/or existing Facts might be deleted.
        try:
            new_fact_or_two, conflicts = mend_fact_timey_wimey(
                controller, new_fact, time_hint,
            )
        except ValueError as err:
            # (lb): I'm very indecisive.
            choices = [
                _("Not so fast!"),
                _("Cannawt!"),
                _("Unpossible!"),
                _("Insidious!"),
                _("Think again!"),
            ]
            msg = _('{} {}').format(random.choice(choices), err)
            controller.client_logger.error(msg)
            dob_in_user_exit(msg)
        return new_fact_or_two, conflicts

    def _prepare_facts(new_fact_or_two):
        edit_facts = []
        orig_facts = []
        new_fact_pk = -1
        for new_fact in new_fact_or_two:
            # If ongoing was squashed, edited_fact.pk > 0, else < 0.
            # There might also be an extended filler gap Fact added.
            if new_fact.pk is None:
                new_fact.pk = new_fact_pk
                new_fact_pk -= 1
                edit_facts.append(new_fact)
            else:
                # The edited fact is in conflicts, and carousel will start on it.
                assert new_fact.pk > 0
        return edit_facts, orig_facts

    def _add_conflicts(conflicts, edit_facts, orig_facts):
        for edited, original in conflicts:
            edit_facts.append(edited)
            # (lb): This is the only place orig_facts is not [edit_fact.copy(), ...].
            orig_facts.append(original)
        edit_fact = edit_facts[0]
        return edit_fact

    def _maybe_prompt_description(edit_fact):
        if not edit_text:
            return

        # User wants to edit description first.
        ask_user_for_edits(
            controller,
            fact=edit_fact,
            always_ask=False,
            restrict_edit='description',
        )

    def _maybe_prompt_actegory(edit_fact):
        if not edit_meta:
            return

        ask_user_for_edits(
            controller,
            fact=edit_fact,
        )

    def _prompt_and_save(edit_facts, orig_facts):
        saved_facts = prompt_and_save_backedup(
            controller,
            edit_facts=edit_facts,
            orig_facts=orig_facts,
            use_carousel=use_carousel,
            yes=yes,
            dry=dry,
            progress=None,
        )
        return saved_facts

    return _add_fact()




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma,  2015-2016 Eric Goller.  All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""Tests hard to reach methods using Carousel."""

import pytest

# (lb): This is the only place in dob project that uses prompt_toolkit
# (which is why it's a requirements/test.pip dependency and not setup.py).
# We need it to fake output to Carousel so we can test a callback method
# that's embedded (out of scope) inside another method (so we cannot call
# directly).
from prompt_toolkit.input.defaults import create_pipe_input
from prompt_toolkit.output import DummyOutput

from dob_viewer.ptkui import re_confirm
from dob_viewer.traverser import save_confirmer as viewer_confirmer

from dob.facts import save_confirmer as proper_confirmer
from dob.facts.import_facts import import_facts

# FIXME/2020-02-01: Could probably simply this test module and use 1 or 2
# Facts from Fact factory -- then moved IMPORT_PATH et al to new module,
# test_import_facts.
IMPORT_PATH = './tests/fixtures/test-import-fixture.rst'
"""Path to the import file fixture, which is full of Factoids."""


# ***

class TestPromptAndSaveBackedUpViaImportFacts(object):
    """Methods to test write_facts_file via the Carousel."""

    def setup_method(self, method):
        self.was_propers = proper_confirmer.prompt_and_save_confirmer
        self.was_viewers = viewer_confirmer.prompt_and_save_confirmer

    def teardown_method(self, method):
        proper_confirmer.prompt_and_save_confirmer = self.was_propers
        viewer_confirmer.prompt_and_save_confirmer = self.was_viewers

    @pytest.mark.parametrize('tres_bools', [
        (True, False, False, ),
        (False, True, False, ),
        (False, False, True, ),
    ])
    def test_prompt_and_save_backedup(
        self,
        controller_with_logging,
        tres_bools,
        mocker,
    ):
        """Tests save_backedup functionality."""

        (backup, use_carousel, dry, ) = tres_bools

        mocker.patch.object(proper_confirmer, 'prompt_and_save_confirmer')
        mocker.patch.object(viewer_confirmer, 'prompt_and_save_confirmer')

        input_stream = open(IMPORT_PATH, 'r')

        # (lb): A somewhat roundabout route to test prompt_and_save_backedup.
        # - Also tests parse_input!
        import_facts(
            controller_with_logging,
            file_in=input_stream,
            use_carousel=use_carousel,
            backup=backup,
            dry=dry,
            # Set 'input' to avoid a guard clause in our code,
            #   "Commands requires user confirmation, or --yes or --dry."
            # The 'input' is usually sent to PTK, but we've mocked that out,
            # so this touch-file, in a sense, is more a slight of hand.
            input=True,
        )


# ***

def _feed_cli_with_input(
    controller_with_logging,
    key_sequence,
    dry,
    mocker,
):
    inp = create_pipe_input()
    input_stream = open(IMPORT_PATH, 'r')
    # Because the key_sequence force-quits, Carousel prompts "Ok?".
    mocker.patch.object(re_confirm, 'confirm', return_value=True)
    try:
        inp.send_text(key_sequence)
        import_facts(
            controller_with_logging,
            file_in=input_stream,
            use_carousel=True,
            # Use backup so backup method actually called!
            backup=True,
            # Testing dry=True adds 1 whole line of coverage!
            dry=dry,
            # Redirect input and output for the test test.
            input=inp,
            output=DummyOutput(),
        )
    finally:
        inp.close()


class TestCarouselBackupCallbackWriteFactsFile(object):
    """Methods to test write_facts_file via the Carousel."""

    @pytest.mark.parametrize(
        ('key_sequence'),
        [
            [
                # Sneak in a test of FactDressed.has_prev_fact:
                '\x1bOD',   # Left arrow ←.
                '\x11',     # Ctrl-Q.
                '\x11',     # Ctrl-Q.
                '\x11',     # Ctrl-Q.
                # We can skip b'' (or '\r') EOL because of the ^Q³.
            ],
        ],
    )
    @pytest.mark.parametrize('dry', [True, False, ])
    def test_carousel_backup_callback_write_facts_file(
        self,
        controller_with_logging,
        key_sequence,
        dry,
        mocker,
    ):
        """Tests write_facts_file via the Carousel, because."""
        # Because the write_facts_file method is embedded in prompt_and_save_backedup,
        # we cannot call it directly to test it. So unless we pull that
        # function out to module scope, or find another way to reach it
        # (perhaps convert prompt_and_save into a class), we can only get
        # at it via Carousel's gallop() -- which we normally wouldn't bother
        # to call from dob, because it's part of a separate component,
        # dob-viewer. So generally we mock out the carousel, but not here.
        # Here, invoke the carousel so that it'll call our callback that we
        # want to test.
        # MAYBE/2020-01-29: What this last comment said.
        # - Maybe make prompt_and_save into a class, or put write_facts_file
        #   at module level, so we can unit test write_facts_file (what we do
        #   here feels more like an integration test).
        _feed_cli_with_input(
            controller_with_logging,
            ''.join(key_sequence),
            dry,
            mocker,
        )




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""Click Group wrapper combines all our custom classes."""

from click_hotoffthehamster_alias import ClickAliasedGroup

from .better_format_usage import ClickBetterUsageGroup
from .better_help_headers import ClickBetterHeadersGroup
from .bunchy_group import ClickBunchyGroup
from .plugin_group import ClickPluginGroup

__all__ = (
    'ClickAliasableBunchyPluginGroup'
)


class ClickAliasableBunchyPluginGroup(
    # ClickBunchyGroup comes first, because
    # its format_commands shadows others'.
    ClickBunchyGroup,
    ClickAliasedGroup,
    ClickPluginGroup,
    # General click.Group overrides (to clean up help output).
    ClickBetterUsageGroup,
    ClickBetterHeadersGroup,
):

    def __init__(self, *args, **kwargs):
        super(ClickAliasableBunchyPluginGroup, self).__init__(*args, **kwargs)

    def command(self, *args, **kwargs):
        # And awaaaaaay we hack!
        # Ensure that all subcommand classes are also this class.
        kwargs.setdefault('cls', self.__class__)
        # So that Click does not complain `Error: Missing command.` when all
        # we want is a little `dob command --help`, always assume
        # invoke_without_command... which is funny, because some commands
        # you can invoke without a command by dropping the option, e.g.,
        # `dob init` works just fine, but `dob init --help` elicits the
        # missing-command complaint from Click. In any case dob.py already
        # wires commands to display help or not as appropriate depending how
        # they're called, e.g., `dob init` runs the init action, but `dob migrate`
        # prints the help for the dob-migrate group of commands.
        kwargs.setdefault('invoke_without_command', True)
        return super(ClickAliasableBunchyPluginGroup, self).command(*args, **kwargs)




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma.  All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

from gettext import gettext as _

import ansiwrap
import click_hotoffthehamster as click
from click_hotoffthehamster.formatting import wrap_text
from click_hotoffthehamster._textwrap import TextWrapper

from dob_bright.termio import attr, click_echo, dob_in_user_exit, fg

from ..clickux.help_strings import NO_ACTIVE_FACT_HELP

__all__ = (
    'echo_fact',
    'echo_latest_ended',
    'echo_ongoing_fact',
    'echo_ongoing_or_ended',
    # Private:
    #  'echo_most_recent',
    #  'echo_single_fact',
)


# ***

def echo_fact(fact):
    click.echo('{}Dry run! New fact{}:\n '.format(
        attr('underlined'),
        attr('reset'),
    ))
    click.echo('{}{}{}{}'.format(
        # FIXME: (lb): Replace hardcoding. Assign from styles.conf. #styling
        fg('steel_blue_1b'),
        attr('bold'),
        fact.friendly_str(description_sep='\n\n'),
        attr('reset'),
    ))


# ***

def echo_latest_ended(controller):
    echo_most_recent(controller, restrict='ended')


def echo_ongoing_fact(controller):
    """
    Return current active Fact.

    Returns:
        None: If everything went alright.

    Raises:
        click.ClickException: If we fail to fetch any active Fact.
    """
    echo_most_recent(
        controller,
        restrict='ongoing',
        empty_msg=NO_ACTIVE_FACT_HELP(controller.ctx),
    )


def echo_ongoing_or_ended(controller):
    echo_most_recent(controller, restrict=None)


def echo_most_recent(controller, restrict=None, empty_msg=None):
    fact = controller.find_latest_fact(restrict=restrict)
    if fact is not None:
        echo_single_fact(controller, fact)
    else:
        empty_msg = empty_msg if empty_msg else _('No facts found.')
        dob_in_user_exit(empty_msg)


# ***

class AnsiWrapper(TextWrapper):

    def __init__(self, *args, **kwargs):
        super(AnsiWrapper, self).__init__(*args, **kwargs)

    def fill(self, *args, **kwargs):
        return ansiwrap.fill(*args, width=self.width, **kwargs)


def echo_single_fact(controller, fact):
    colorful = controller.config['term.use_color']
    localize = controller.config['time.tz_aware']
    friendly = fact.friendly_str(
        shellify=False,
        description_sep=': ',
        localize=localize,
        colorful=colorful,
        show_elapsed=True,
    )
    # Click's default wrap_text behavior uses Click TextWrapper class, which
    # extends Python's textwrap.TextWrapper, which is not ANSI-aware. So we
    # extent TextWrapper to redirect it to ansiwrap, which is ANSI-couth.
    # FIXME/2019-11-22: (lb): Make this width CONFIGable.
    wrapped = wrap_text(friendly, width=100, preserve_paragraphs=True, cls=AnsiWrapper)
    click_echo(wrapped)


# ***

def write_fact_block_format(fact_f, fact, rule, is_first_fact):
    write_fact_separator(fact_f, rule, is_first_fact)
    friendly_str = fact.friendly_str(
        # description_sep='\n\n',
        description_sep=': ',
        shellify=False,
        colorful=False,
        empty_actegory_placeholder='',
    )
    fact_f.write(friendly_str)


def write_fact_separator(fact_f, rule, is_first_fact):
    RULE_WIDTH = 76  # How wide to print the between-facts separator.
    if is_first_fact:
        return
    fact_f.write('\n\n')
    if rule:
        fact_f.write('{}\n\n'.format(rule * RULE_WIDTH))




# This file exists within 'dob-bright':
#
#   https://github.com/hotoffthehamster/dob-bright
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

import pytest

from unittest.mock import PropertyMock

from nark.tests.conftest import *  # noqa: F401, F403
from nark.tests.backends.sqlalchemy.conftest import *  # noqa: F401, F403

from nark.backends.sqlalchemy import objects
from nark.backends.sqlalchemy import storage
from nark.backends.sqlalchemy.managers.migrate import MigrationsManager
from nark.control import NarkControl

from dob_bright.config.urable import ConfigUrable

# (lb): Note that setting pytest_plugins = (...) won't work from this file
# (because not a conftest.py module, I'd guess). One options is to *-import
# necessary files, e.g.,
#   from nark.tests.backends.sqlalchemy.conftest import *  # noqa: F401, F403
# Another option is to add the necessary fixture modules to the pytest_plugins
# list in this directory's conftest.py file, which seems like the better option.
# - tl;dr: Use pytest_plugins, not import, to source additional fixtures.


@pytest.fixture
def dob_runner(alchemy_store, test_fact_cls, controller, mocker, runner):
    """Provide a convenient fixture for running CLI commands against inited dob."""

    # Get the SQLAlchemy Engine() created by alchemy_store,
    # and stashed in the Alchemy MetaData() class maintained
    # by the nark.backends.sqlalchemy.objects module.
    engine = objects.metadata.bind

    # Add the migration version, otherwise dob will complain db not ready.
    migrations = MigrationsManager(alchemy_store)
    migrations.control(version=None, engine=engine)

    mocker.patch.object(
        storage.SQLAlchemyStore, 'create_storage_engine', return_value=engine,
    )

    mocker.patch.object(NarkControl, '_get_store', return_value=alchemy_store)

    mocker.patch.object(
        ConfigUrable, 'config_root', return_value=controller.config,
        new_callable=PropertyMock,
    )

    cfgfile_exists = True
    mocker.patch.object(
        ConfigUrable, 'load_configfile', return_value=cfgfile_exists,
    )

    return runner




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma,  2015-2016 Eric Goller.  All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

from dob.dob import _license


class TestDobLicenseWrapper(object):
    """Unittests for ``license`` command."""

    def test_dob_license_wrapper(self, capsys):
        """Make sure the license text is actually displayed."""
        _license()
        out, err = capsys.readouterr()
        assert out.startswith("GNU GENERAL PUBLIC LICENSE")
        assert "Version 3, 29 June 2007" in out




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma,  2015-2016 Eric Goller.  All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

from dob import __package_name__, get_version

from dob.details import echo_app_details


class TestDetails(object):
    """Unittests for the ``details`` command."""

    def test_details_general_data_is_shown(self, controller, capsys):
        """Make sure user recieves the desired output."""
        controller.setup_tty_color(use_color=False)
        echo_app_details(controller)
        out, err = capsys.readouterr()
        startswiths = (
            'You are running {} version {}'.format(
                __package_name__, get_version(),
            ),
            'Configuration file at: ',
            'Plugins directory at: ',
            'Logfile stored at: ',
            'Using sqlite on database: :memory:',
        )
        for idx, line in enumerate(out.splitlines()):
            assert line.startswith(startswiths[idx])

    def test_details_sqlite(self, controller, appdirs, mocker, capsys):
        """Make sure database details for sqlite are shown properly."""
        mocker.patch.object(controller, '_get_store')
        engine, path = 'sqlite', appdirs.user_data_dir
        controller.config['db.engine'] = engine
        controller.config['db.path'] = path
        echo_app_details(controller)
        out, err = capsys.readouterr()
        for item in (engine, path):
            assert item in out
        assert out.splitlines()[-1] == 'Using {} on database: {}'.format(engine, path)

    def test_details_non_sqlite(
        self,
        controller,
        capsys,
        db_port,
        db_host,
        db_name,
        db_user,
        db_password,
        mocker,
    ):
        """
        Make sure database details for non-sqlite are shown properly.

        We need to mock the backend Controller because it would try to setup a
        database connection right away otherwise.
        """
        mocker.patch.object(controller, '_get_store')
        controller.config['db.engine'] = 'postgres'
        controller.config['db.name'] = db_name
        controller.config['db.host'] = db_host
        controller.config['db.user'] = db_user
        controller.config['db.password'] = db_password
        controller.config['db.port'] = db_port
        echo_app_details(controller)
        out, err = capsys.readouterr()
        for item in ('postgres', db_host, db_name, db_user):
            assert item in out
        if db_port:
            assert db_port in out
        assert db_password not in out




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""Module to define command group decorators and section headers."""

from gettext import gettext as _

from .help_header import help_header_format
from ..run_cli import run

__all__ = (
    'cmd_bunch_group_add_fact',
    'cmd_bunch_group_dbms',
    'cmd_bunch_group_edit',
    'cmd_bunch_group_generate_report',
    'cmd_bunch_group_get_meta',
    'cmd_bunch_group_introducing',
    'cmd_bunch_group_ongoing_fact',
    'cmd_bunch_group_personalize',
    'cmd_bunch_group_plugin',
)


# ***
# *** Help command group headers.
# ***

def help_header_introducing():
    return help_header_format(_('Learn and Setup Dob'))


def help_header_edit():
    return help_header_format(_('Run the Editor'))


def help_header_personalize():
    return help_header_format(_('Personalize the Editor'))


def help_header_get_meta():
    return help_header_format(_('Manage the Application'))


def help_header_generate_report():
    return help_header_format(_('Generate Reports'))


def help_header_dbms():
    return help_header_format(_('Manage the Database'))


def help_header_add_fact():
    return help_header_format(_('Add Facts from the CLI'))


def help_header_ongoing_fact():
    return help_header_format(_('Work on the Latest Fact'))


def help_header_plugin():
    return help_header_format(_('Plugin Commands'))


# ***
# *** Help command group headers.
# ***

def cmd_bunch_group_introducing(cmd):
    run.add_to_bunch(cmd, help_header_introducing, 100)
    return cmd


def cmd_bunch_group_edit(cmd):
    run.add_to_bunch(cmd, help_header_edit, 200)
    return cmd


def cmd_bunch_group_personalize(cmd):
    run.add_to_bunch(cmd, help_header_personalize, 600)
    return cmd


def cmd_bunch_group_get_meta(cmd):
    run.add_to_bunch(cmd, help_header_get_meta, 400)
    return cmd


def cmd_bunch_group_generate_report(cmd):
    run.add_to_bunch(cmd, help_header_generate_report, 300)
    return cmd


def cmd_bunch_group_dbms(cmd):
    run.add_to_bunch(cmd, help_header_dbms, 500)
    return cmd


def cmd_bunch_group_add_fact(cmd):
    run.add_to_bunch(cmd, help_header_add_fact, 700)
    return cmd


def cmd_bunch_group_ongoing_fact(cmd):
    run.add_to_bunch(cmd, help_header_ongoing_fact, 710)
    return cmd


def cmd_bunch_group_plugin(cmd):
    run.add_to_bunch(cmd, help_header_plugin, 999)
    return cmd

